"""
Unit tests for Instructor Tool block
"""
import unittest

import ddt
from mock import Mock, patch
from xblock.field_data import DictFieldData

from problem_builder.instructor_tool import (COURSE_BLOCKS_API,
                                             InstructorToolBlock)


@ddt.ddt
class TestInstructorToolBlock(unittest.TestCase):
    """
    Test InstructorToolBlock with some mocked data.
    """

    def _get_block(self, block_info):
        # Build spec
        spec = ['scope_ids', 'runtime', 'has_children', 'children']
        for attr in block_info.get('attrs', []):
            spec.append(attr)
        # Assemble block
        block = Mock(spec=spec)
        scope_ids_mock = Mock()
        scope_ids_mock.usage_id = block_info.get('usage_id')
        block.scope_ids = scope_ids_mock
        block.runtime = self.runtime_mock
        block.children = []
        for attr, val in block_info.get('attrs', {}).items():
            setattr(block, attr, val)
        return block

    def setUp(self):
        self.course_id = 'course-v1:edX+DemoX+Demo_Course'
        self.runtime_mock = Mock()
        self.runtime_mock.get_block = self._get_block
        self.runtime_mock.course_id = self.course_id
        scope_ids_mock = Mock()
        scope_ids_mock.usage_id = u'0'
        self.block = InstructorToolBlock(
            self.runtime_mock, field_data=DictFieldData({}), scope_ids=scope_ids_mock
        )

    def test_student_view_template_args(self):
        """
        Check if `student_view` calls rendering method of template loader
        with correct arguments.
        """
        block_choices = {
            'Multiple Choice Question': 'MCQBlock',
            'Multiple Response Question': 'MRQBlock',
            'Rating Question': 'RatingBlock',
            'Long Answer': 'AnswerBlock',
        }

        with patch('problem_builder.instructor_tool.loader') as patched_loader:
            patched_loader.render_template.return_value = u''
            self.block.student_view()
            patched_loader.render_template.assert_called_once_with('templates/html/instructor_tool.html', {
                'block_choices': block_choices,
                'course_blocks_api': COURSE_BLOCKS_API,
                'root_block_id': self.course_id,
            })

    def test_author_view(self):
        """
        Check if author view shows appropriate message when viewing units
        containing Instructor Tool block in Studio.
        """
        fragment = self.block.author_view()
        self.assertIn('This block only works from the LMS.', fragment.content)

    def test_studio_view(self):
        """
        Check if studio view is present and shows appropriate message when
        trying to edit Instructor Tool block in Studio.
        """
        try:
            fragment = self.block.studio_view()
        except AttributeError:
            self.fail('Studio view not defined.')
        self.assertIn('This is a preconfigured block. It is not editable.', fragment.content)



"""
Django settings for the problem-builder project.
For more information on this file, see
https://docs.djangoproject.com/en/1.11/topics/settings/
For the full list of settings and their values, see
https://docs.djangoproject.com/en/1.11/ref/settings/
"""

# Build paths inside the project like this: os.path.join(BASE_DIR, ...)
import os

import yaml

BASE_DIR = os.path.dirname(os.path.dirname(__file__))


# Quick-start development settings - unsuitable for production
# See https://docs.djangoproject.com/en/1.11/howto/deployment/checklist/

# SECURITY WARNING: keep the secret key used in production secret!
# This is just a container for running tests, it's okay to allow it to be
# defaulted here if not present in environment settings
SECRET_KEY = os.environ.get('SECRET_KEY', '",cB3Jr.?xu[x_Ci]!%HP>#^AVmWi@r/W3u,w?pY+~J!R>;WN+,3}Sb{K=Jp~;&k')

# SECURITY WARNING: don't run with debug turned on in production!
# This is just a container for running tests
DEBUG = True

TEMPLATE_DEBUG = True

ALLOWED_HOSTS = []


# Application definition

INSTALLED_APPS = (
    'statici18n',
    'problem_builder',
    'django.contrib.auth',
    'django.contrib.contenttypes',
)

# Internationalization
# https://docs.djangoproject.com/en/1.6/topics/i18n/

LANGUAGE_CODE = 'en-us'

TIME_ZONE = 'UTC'

USE_I18N = True

USE_L10N = True

USE_TZ = True


# Static files (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/1.6/howto/static-files/

STATIC_URL = '/static/'

LOCALE_PATHS = [
    os.path.join(BASE_DIR, 'problem_builder/translations'),
]

# statici18n
# http://django-statici18n.readthedocs.io/en/latest/settings.html

with open(os.path.join(BASE_DIR, 'problem_builder/translations/config.yaml'), 'r') as locale_config_file:
    locale_config = yaml.load(locale_config_file)

    LANGUAGES = [
        (code, code,)
        for code in locale_config['locales'] + locale_config['dummy_locales']
    ]

STATICI18N_DOMAIN = 'textjs'

# `ProblemI18N` might collide with other modules, using `ProblemBuilderXBlockI18N` to be more explicit
STATICI18N_NAMESPACE = 'ProblemBuilderXBlockI18N'
STATICI18N_PACKAGES = (
    'problem_builder',
)
STATICI18N_ROOT = 'problem_builder/public/js'
STATICI18N_OUTPUT_DIR = 'translations'



# -*- coding: utf-8 -*-
#
# Copyright (c) 2014-2015 Harvard, edX & OpenCraft
#
# This software's license gives you freedom; you can copy, convey,
# propagate, redistribute and/or modify this program under the terms of
# the GNU Affero General Public License (AGPL) as published by the Free
# Software Foundation (FSF), either version 3 of the License, or (at your
# option) any later version of the AGPL published by the FSF.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero
# General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program in a file in the toplevel directory called
# "AGPLv3".  If not, see <http://www.gnu.org/licenses/>.
#

# Imports ###########################################################

import logging
import uuid

import six
from xblock.core import XBlock
from xblock.fields import Float, Scope, String
from xblock.fragment import Fragment
from xblockutils.resources import ResourceLoader
from xblockutils.studio_editable import StudioEditableXBlockMixin

from .mixins import (QuestionMixin, StudentViewUserStateMixin,
                     XBlockWithTranslationServiceMixin)
from .sub_api import SubmittingXBlockMixin, sub_api

# Globals ###########################################################

log = logging.getLogger(__name__)
loader = ResourceLoader(__name__)


# Make '_' a no-op so we can scrape strings
def _(text):
    return text

# Classes ###########################################################


@XBlock.needs("i18n")
class SliderBlock(
    SubmittingXBlockMixin, QuestionMixin, StudioEditableXBlockMixin, XBlockWithTranslationServiceMixin,
    StudentViewUserStateMixin, XBlock,
):
    """
    An XBlock used by students to indicate a numeric value on a sliding scale.
    The student's answer is always considered "correct".
    """
    CATEGORY = 'pb-slider'
    STUDIO_LABEL = _(u"Ranged Value Slider")
    USER_STATE_FIELDS = ['student_value']

    answerable = True

    min_label = String(
        display_name=_("Low"),
        help=_("Label for low end of the range"),
        scope=Scope.content,
        default=_("0%"),
    )
    max_label = String(
        display_name=_("High"),
        help=_("Label for high end of the range"),
        scope=Scope.content,
        default=_("100%"),
    )

    question = String(
        display_name=_("Question"),
        help=_("Question to ask the student (optional)"),
        scope=Scope.content,
        default="",
        multiline_editor=True,
    )

    student_value = Float(
        # The value selected by the student
        default=None,
        scope=Scope.user_state,
    )

    editable_fields = ('min_label', 'max_label', 'display_name', 'question', 'show_title')

    @property
    def url_name(self):
        """
        Get the url_name for this block. In Studio/LMS it is provided by a mixin, so we just
        defer to super(). In the workbench or any other platform, we use the name.
        """
        try:
            return super(SliderBlock, self).url_name
        except AttributeError:
            return self.name

    def mentoring_view(self, context):
        """ Main view of this block """
        context = context.copy() if context else {}
        context['question'] = self.question
        context['slider_id'] = 'pb-slider-{}'.format(uuid.uuid4().hex[:20])
        context['initial_value'] = int(self.student_value*100) if self.student_value is not None else 50
        context['min_label'] = self.min_label
        context['max_label'] = self.max_label
        context['title'] = self.display_name_with_default
        context['hide_header'] = context.get('hide_header', False) or not self.show_title
        context['instructions_string'] = self._("Select a value from {min_label} to {max_label}").format(
            min_label=self.min_label, max_label=self.max_label
        )
        html = loader.render_template('templates/html/slider.html', context)

        fragment = Fragment(html)
        fragment.add_javascript_url(self.runtime.local_resource_url(self, 'public/js/slider.js'))
        fragment.initialize_js('SliderBlock')
        return fragment

    student_view = mentoring_view
    preview_view = mentoring_view

    def student_view_data(self, context=None):
        return {
            'id': self.name,
            'block_id': six.text_type(self.scope_ids.usage_id),
            'display_name': self.display_name_with_default,
            'type': self.CATEGORY,
            'question': self.question,
            'min_label': self.min_label,
            'max_label': self.max_label,
            'title': self.display_name_with_default,
            'hide_header': not self.show_title,
        }

    def author_view(self, context):
        """
        Add some HTML to the author view that allows authors to see the ID of the block, so they
        can refer to it in other blocks such as Plot blocks.
        """
        context['hide_header'] = True  # Header is already shown in the Studio wrapper
        fragment = self.student_view(context)
        fragment.add_content(loader.render_template('templates/html/slider_edit_footer.html', {
            "url_name": self.url_name
        }))
        return fragment

    def get_last_result(self):
        """ Return the current/last result in the required format """
        if self.student_value is None:
            return {}
        return {
            'submission': self.student_value,
            'status': 'correct',
            'tips': [],
            'weight': self.weight,
            'score': 1,
        }

    def get_results(self, _previous_result_unused=None):
        """ Alias for get_last_result() """
        return self.get_last_result()

    def submit(self, value):
        log.debug(u'Received Slider submission: "%s"', value)
        if value < 0 or value > 1:
            return {}  # Invalid
        self.student_value = value
        if sub_api:
            # Also send to the submissions API:
            sub_api.create_submission(self.student_item_key, value)
        result = self.get_last_result()
        log.debug(u'Slider submission result: %s', result)
        return result

    def get_submission_display(self, submission):
        """
        Get the human-readable version of a submission value
        """
        return submission * 100

    def validate_field_data(self, validation, data):
        """
        Validate this block's field data.
        """
        super(SliderBlock, self).validate_field_data(validation, data)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.4 on 2017-09-19 13:57
from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('problem_builder', '0005_auto_20170112_1021'),
    ]

    operations = [
        migrations.AlterUniqueTogether(
            name='answer',
            unique_together=set([('student_id', 'course_key', 'name')]),
        ),
        migrations.RemoveField(
            model_name='answer',
            name='course_id',
        ),
    ]



# -*- coding: utf-8 -*-
#
# Copyright (c) 2014-2015 Harvard, edX & OpenCraft
#
# This software's license gives you freedom; you can copy, convey,
# propagate, redistribute and/or modify this program under the terms of
# the GNU Affero General Public License (AGPL) as published by the Free
# Software Foundation (FSF), either version 3 of the License, or (at your
# option) any later version of the AGPL published by the FSF.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero
# General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program in a file in the toplevel directory called
# "AGPLv3".  If not, see <http://www.gnu.org/licenses/>.
#

# Imports ###########################################################

from django.contrib.auth.models import User
from django.db import models
from django.db.models.signals import pre_delete

try:
    # workaround so we don't explicitly import the AnonymousUserId model from LMS
    from student.models import AnonymousUserId
except ImportError:
    AnonymousUserId = None


# Classes ###########################################################

class Answer(models.Model):
    """
    Django model used to store AnswerBlock data that need to be shared
    and queried accross XBlock instances (workaround).

    TODO: Deprecate this and move to edx-submissions
    """

    class Meta:
        # Since problem_builder isn't added to INSTALLED_APPS until it's imported,
        # specify the app_label here.
        app_label = 'problem_builder'
        unique_together = (
            ('student_id', 'course_key', 'name'),
        )

    name = models.CharField(max_length=50, db_index=True)
    student_id = models.CharField(max_length=50, db_index=True)
    course_key = models.CharField(max_length=255, db_index=True)
    student_input = models.TextField(blank=True, default=u'')
    created_on = models.DateTimeField(u'created on', auto_now_add=True)
    modified_on = models.DateTimeField(u'modified on', auto_now=True)

    def save(self, *args, **kwargs):
        # Force validation of max_length
        self.full_clean()
        super(Answer, self).save(*args, **kwargs)


class Share(models.Model):
    """
    The XBlock User Service does not permit XBlocks instantiated with non-staff users
    to query for arbitrary anonymous user IDs. In order to make sharing work, we have
    to store them here.
    """
    shared_by = models.ForeignKey(User, related_name='problem_builder_shared_by')
    submission_uid = models.CharField(max_length=32)
    block_id = models.CharField(max_length=255, db_index=True)
    shared_with = models.ForeignKey(User, related_name='problem_builder_shared_with')
    notified = models.BooleanField(default=False, db_index=True)

    class Meta(object):
        # Since problem_builder isn't added to INSTALLED_APPS until it's imported,
        # specify the app_label here.
        app_label = 'problem_builder'
        unique_together = (('shared_by', 'shared_with', 'block_id'),)


# Signals ###########################################################

def delete_anonymous_user_answers(sender, **kwargs):
    """
    Delete Answer records when an AnonymousUserId is deleted.
    """
    instance = kwargs['instance']
    Answer.objects.filter(student_id=instance.anonymous_user_id).delete()


if AnonymousUserId:
    pre_delete.connect(delete_anonymous_user_answers, sender=AnonymousUserId)



# -*- coding: utf-8 -*-
#
# Copyright (c) 2014-2015 Harvard, edX & OpenCraft
#
# This software's license gives you freedom; you can copy, convey,
# propagate, redistribute and/or modify this program under the terms of
# the GNU Affero General Public License (AGPL) as published by the Free
# Software Foundation (FSF), either version 3 of the License, or (at your
# option) any later version of the AGPL published by the FSF.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero
# General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program in a file in the toplevel directory called
# "AGPLv3".  If not, see <http://www.gnu.org/licenses/>.
#

# Imports ###########################################################
from .base_test import MentoringBaseTest

# Classes ###########################################################


class MentoringProgressionTest(MentoringBaseTest):

    def assert_warning(self, warning_dom, link_href):
        """
        Check that the provided DOM element is a progression warning, and includes a link with a href
        pointing to `link_href`
        """
        self.assertEqual(warning_dom.text, 'You need to complete the previous step before attempting this step.')
        warning_link = warning_dom.find_element_by_xpath('./*')
        self.assertTrue(
            warning_link.get_attribute('href').endswith(link_href),
            "expected link href '{}' to end with '{}'".format(warning_link.get_attribute('href'), link_href)
        )

    def assert_warning_is_hidden(self, mentoring):
        for elem in mentoring.find_elements_by_css_selector('.warning'):
            self.assertFalse(elem.is_displayed())

    def test_progression(self):
        """
        Mentoring blocks after the current step in the workflow should redirect user to current step
        """
        # Initial - Step 1 ok, steps 2&3 redirect to step 1
        mentoring = self.go_to_page('Progression 1')
        self.assert_warning_is_hidden(mentoring)
        submit = mentoring.find_element_by_css_selector('.submit input.input-main')
        self.assertFalse(submit.is_enabled())

        mentoring = self.go_to_page('Progression 2')
        warning = mentoring.find_element_by_css_selector('.warning')
        self.assert_warning(warning, '/jump_to_id/mentoring_first')
        submit = mentoring.find_element_by_css_selector('.submit input.input-main')
        self.assertFalse(submit.is_enabled())

        mentoring = self.go_to_page('Progression 3')
        warning = mentoring.find_element_by_css_selector('.warning')
        self.assert_warning(warning, '/jump_to_id/mentoring_first')
        submit = mentoring.find_element_by_css_selector('.submit input.input-main')
        self.assertFalse(submit.is_enabled())

        # Should be impossible to complete step 2
        mentoring = self.go_to_page('Progression 2')
        answer = mentoring.find_element_by_css_selector('textarea')
        answer.send_keys('This is the answer')
        submit = mentoring.find_element_by_css_selector('.submit input.input-main')
        self.assertTrue(submit.is_displayed() and submit.is_enabled())
        submit.click()
        self.wait_until_disabled(submit)

        messages = mentoring.find_element_by_css_selector('.messages')
        self.assertTrue(messages.is_displayed())
        self.assertIn(
            'You need to complete all previous steps before being able to complete the current one.',
            messages.text)

        mentoring = self.go_to_page('Progression 2')
        warning = mentoring.find_element_by_css_selector('.warning')
        self.assert_warning(warning, '/jump_to_id/mentoring_first')

        mentoring = self.go_to_page('Progression 3')
        warning = mentoring.find_element_by_css_selector('.warning')
        self.assert_warning(warning, '/jump_to_id/mentoring_first')

        # Complete step 1 - now only step 3 should redirect, and to step 2
        mentoring = self.go_to_page('Progression 1')
        answer = mentoring.find_element_by_css_selector('textarea')
        answer.send_keys('This is the answer')
        submit = mentoring.find_element_by_css_selector('.submit input.input-main')
        self.assertTrue(submit.is_displayed() and submit.is_enabled())
        submit.click()
        self.wait_until_disabled(submit)
        self.assert_warning_is_hidden(mentoring)

        messages = mentoring.find_element_by_css_selector('.messages')
        self.assertFalse(messages.is_displayed())

        mentoring = self.go_to_page('Progression 3')
        warning = mentoring.find_element_by_css_selector('.warning')
        self.assert_warning(warning, '/jump_to_id/progression_2')

        mentoring = self.go_to_page('Progression 2')
        self.assert_warning_is_hidden(mentoring)

        # Complete step 2 - no more warnings anywhere
        submit = mentoring.find_element_by_css_selector('.submit input.input-main')
        answer = mentoring.find_element_by_css_selector('textarea')
        self.assertEqual(answer.text, "")  # Earlier attempt to submit did not save
        answer.send_keys('This is the answer')
        self.assertTrue(submit.is_displayed() and submit.is_enabled())
        submit.click()
        self.wait_until_disabled(submit)

        messages = mentoring.find_element_by_css_selector('.messages')
        self.assertFalse(messages.is_displayed())

        mentoring = self.go_to_page('Progression 1')
        self.assert_warning_is_hidden(mentoring)

        mentoring = self.go_to_page('Progression 2')
        self.assert_warning_is_hidden(mentoring)

        mentoring = self.go_to_page('Progression 3')
        self.assert_warning_is_hidden(mentoring)

        # Should be able to complete step 3 too now
        answer = mentoring.find_element_by_css_selector('textarea')
        answer.send_keys('This is the answer')
        submit = mentoring.find_element_by_css_selector('.submit input.input-main')
        submit.click()
        self.wait_until_disabled(submit)

        messages = mentoring.find_element_by_css_selector('.messages')
        self.assertFalse(messages.is_displayed())



"""
If an author makes changes to the block after students have started using it, will bad things
happen?
"""
import time

from .base_test import ProblemBuilderBaseTest


class AuthorChangesTest(ProblemBuilderBaseTest):
    """
    Test various scenarios involving author changes made to a block already in use by students
    """
    def setUp(self):
        super(AuthorChangesTest, self).setUp()
        self.load_scenario("author_changes.xml", {}, load_immediately=False)
        self.refresh_page()

    def refresh_page(self):
        """
        [Re]load the page with our scenario
        """
        self.pb_block_dom = self.go_to_view("student_view")
        # At this point the ajax request that initializes the Mentoring block
        # might be still in progres. Race conditions resulting in duplicate field data
        # can occur if we try to reload the block at the same time.
        # Sleep 200ms to wait for the ajax request to finish, unfortunately I wasn't
        # able to find a better way.
        time.sleep(0.2)
        self.reload_pb_block()

    def reload_pb_block(self):
        """
        [Re]load the Problem Builder block, potentially with updated field data
        """
        vertical = self.load_root_xblock()
        self.pb_block = vertical.runtime.get_block(vertical.children[0])

    def submit_answers(self, q1_answer='yes', q2_answer='elegance', q3_answer="It's boring."):
        """ Answer all three questions in the 'author_changes.xml' scenario correctly """
        self.pb_block_dom.find_element_by_css_selector('input[name=q1][value={}]'.format(q1_answer)).click()
        self.pb_block_dom.find_element_by_css_selector('input[name=q2][value={}]'.format(q2_answer)).click()
        self.pb_block_dom.find_element_by_css_selector('textarea').send_keys(q3_answer)
        self.click_submit(self.pb_block_dom)

    def test_delete_question(self):
        """ Test what the block behaves correctly when deleting a question """
        # First, submit an answer to each of the three questions, but get the second question wrong:
        self.submit_answers(q2_answer='bugs')
        self.reload_pb_block()
        self.assertEqual(self.pb_block.score.percentage, 67)

        # Delete the second question:
        self.pb_block.children = [self.pb_block.children[0], self.pb_block.children[2]]
        self.pb_block.save()
        self.reload_pb_block()

        # Now that the wrong question is deleted, the student should have a perfect score:
        self.assertEqual(self.pb_block.score.percentage, 100)
        # NOTE: This is questionable, since the block does not send a new 'grade' event to the
        # LMS. So the LMS 'grade' (based on the event sent when the student actually submitted
        # the answers) and the block's current 'score' may be different.

    def test_reweight_question(self):
        """ Test what the block behaves correctly when changing the weight of a question """
        # First, submit an answer to each of the three questions, but get the first question wrong:
        self.submit_answers(q1_answer='no')
        self.reload_pb_block()
        self.assertEqual(self.pb_block.score.percentage, 67)

        # Re-weight Q1 to '5':
        q1 = self.pb_block.runtime.get_block(self.pb_block.children[0])
        q1.weight = 5
        q1.save()
        self.reload_pb_block()
        self.assertEqual(self.pb_block.score.percentage, 29)  # 29% is 2 out of 7 (5+1+1)

        # Delete Q2 (the MRQ)
        self.pb_block.children = [self.pb_block.children[0], self.pb_block.children[2]]
        self.pb_block.save()
        self.reload_pb_block()

        # Now, the student's score should be 1 out of 6 (only q3 is correct):
        self.assertEqual(self.pb_block.score.percentage, 17)



"""
Tests common to Problem Builder and Step Builder
"""
import unittest

import ddt
from xblock.core import XBlock

from problem_builder.mentoring import (MentoringBlock,
                                       MentoringWithExplicitStepsBlock)

from .utils import ScoresTestMixin, instantiate_block


@ddt.ddt
class TestBuilderBlocks(ScoresTestMixin, unittest.TestCase):
    """ Unit tests for Problem Builder and Step Builder """

    @ddt.data(MentoringBlock, MentoringWithExplicitStepsBlock)
    def test_interface(self, block_cls):
        """
        Basic tests of the block's public interface.
        """
        self.assertTrue(issubclass(block_cls, XBlock))
        self.assertTrue(block_cls.has_children)

        block = instantiate_block(block_cls)
        self.assertTrue(block.has_children)
        self.assert_produces_scores(block)



# -*- coding: utf-8 -*-
#
# Copyright (c) 2014-2015 Harvard, edX & OpenCraft
#
# This software's license gives you freedom; you can copy, convey,
# propagate, redistribute and/or modify this program under the terms of
# the GNU Affero General Public License (AGPL) as published by the Free
# Software Foundation (FSF), either version 3 of the License, or (at your
# option) any later version of the AGPL published by the FSF.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero
# General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program in a file in the toplevel directory called
# "AGPLv3".  If not, see <http://www.gnu.org/licenses/>.
#

# Imports ###########################################################

from .base_test import (GetChoices, MentoringAssessmentBaseTest,
                        ProblemBuilderBaseTest)

# Classes ###########################################################


class SliderBlockTestMixins(object):
    """ Mixins for testing slider blocks. Assumes only one slider block is on the page. """

    def get_slider_value(self):
        return int(self.browser.execute_script("return $('.pb-slider-range').val()"))

    def set_slider_value(self, val):
        self.browser.execute_script("$('.pb-slider-range').val(arguments[0]).change()", val)


class SliderBlockTest(SliderBlockTestMixins, ProblemBuilderBaseTest):
    """
    Tests for the SliderBlock inside a normal Problem Builder block.
    """
    def test_simple_flow(self):
        """ Test a regular Problem Builder block containing one slider """
        pb_wrapper = self.load_scenario("slider_problem.xml", {"include_mcq": False})
        self.wait_for_init()
        # The initial value should be 50 and submit should be enabled since 50 is a valid value:
        self.assertEqual(self.get_slider_value(), 50)
        self.expect_checkmark_visible(False)
        self.expect_submit_enabled(True)
        # Set the value to 75:
        self.set_slider_value(75)
        self.assertEqual(self.get_slider_value(), 75)
        self.click_submit(pb_wrapper)
        # Now, we expect submit to be disabled and the checkmark to be visible:
        self.expect_checkmark_visible(True)
        self.expect_submit_enabled(False)
        # Now change the value, and the button/checkmark should reset:
        self.set_slider_value(45)
        self.expect_checkmark_visible(False)
        self.expect_submit_enabled(True)
        # Now reload the page:
        pb_wrapper = self.reload_page()
        self.wait_for_init()
        # Now the initial value should be 75 and submit should be disabled (to discourage submitting the same answer):
        self.assertEqual(self.get_slider_value(), 75)
        self.wait_until_visible(self.checkmark)
        self.expect_checkmark_visible(True)
        self.expect_submit_enabled(False)

    def test_simple_flow_with_peer(self):
        """ Test a regular Problem Builder block containing one slider and an MCQ """
        pb_wrapper = self.load_scenario("slider_problem.xml", {"include_mcq": True})
        self.wait_for_init()
        # The initial value should be 50 and submit should be disabled until an MCQ choice is selected
        self.assertEqual(self.get_slider_value(), 50)
        self.expect_checkmark_visible(False)
        self.expect_submit_enabled(False)
        # Set the value to 15:
        self.set_slider_value(15)
        self.assertEqual(self.get_slider_value(), 15)
        self.expect_submit_enabled(False)
        # Choose a choice:
        GetChoices(pb_wrapper).select('Yes')
        self.expect_submit_enabled(True)
        self.click_submit(pb_wrapper)
        # Now, we expect submit to be disabled and the checkmark to be visible:
        self.expect_checkmark_visible(True)
        self.expect_submit_enabled(False)
        # Now change the value, and the button/checkmark should reset:
        self.set_slider_value(20)
        self.expect_checkmark_visible(False)
        self.expect_submit_enabled(True)


class SliderStepBlockTest(SliderBlockTestMixins, MentoringAssessmentBaseTest):
    """
    Tests for the SliderBlock inside a Step Builder block.
    """

    def test_step_with_slider(self):
        """ Test a regular Step Builder block containing one slider and an MCQ """
        step_builder, controls = self.load_assessment_scenario("slider_step.xml")
        self.wait_for_init()
        self.assertEqual(self.get_slider_value(), 50)

        # Check step 1 (the slider step):
        question = self.expect_question_visible(1, step_builder, question_text="Information Reliability")
        self.assertIn("How reliable is this information?", question.text)
        self.assertIn("Select a value from 0% to 100%", question.text)  # Screen reader explanation
        self.assertTrue(controls.submit.is_enabled())
        self.assert_hidden(controls.try_again)

        self.set_slider_value(0)
        controls.submit.click()
        self.do_submit_wait(controls, last=False)
        self.wait_until_clickable(controls.next_question)
        controls.next_question.click()

        # Submit step 2:
        question = self.expect_question_visible(2, step_builder)
        GetChoices(question).select("Yes")
        controls.submit.click()
        self.do_submit_wait(controls, last=True)
        self.wait_until_clickable(controls.review)

        controls.review.click()
        self.wait_until_visible(controls.try_again)
        # You can't get a slider question wrong, but it does count as one correct point by default:
        self.assertIn("You answered 2 questions correctly", step_builder.text)



import unittest

from mock import Mock
from xblock.field_data import DictFieldData

from problem_builder.mixins import QuestionMixin, StepParentMixin
from problem_builder.step import MentoringStepBlock

from .utils import BlockWithChildrenTestMixin


class Parent(StepParentMixin):
    @property
    def children(self):
        """ Return an ID for each of our chilren"""
        return range(0, len(self._children))

    @property
    def runtime(self):
        return Mock(
            get_block=lambda i: self._children[i],
            load_block_type=lambda i: type(self._children[i]),
            id_reader=Mock(get_definition_id=lambda i: i, get_block_type=lambda i: i)
        )

    def _set_children_for_test(self, *children):
        self._children = children
        for idx, child in enumerate(self._children):
            try:
                child.get_parent = lambda: self
                child.scope_ids = Mock(usage_id=idx)
            except AttributeError:
                pass


class Step(QuestionMixin):
    def __init__(self):
        pass


class NotAStep(object):
    pass


class TestQuestionMixin(unittest.TestCase):
    def test_single_step_is_returned_correctly(self):
        block = Parent()
        step = Step()
        block._children = [step]

        steps = [block.runtime.get_block(cid) for cid in block.step_ids]
        self.assertSequenceEqual(steps, [step])

    def test_only_steps_are_returned(self):
        block = Parent()
        step1 = Step()
        step2 = Step()
        block._set_children_for_test(step1, 1, "2", "Step", NotAStep(), False, step2, NotAStep())

        steps = [block.runtime.get_block(cid) for cid in block.step_ids]
        self.assertSequenceEqual(steps, [step1, step2])

    def test_proper_number_is_returned_for_step(self):
        block = Parent()
        step1 = Step()
        step2 = Step()
        block._set_children_for_test(step1, 1, "2", "Step", NotAStep(), False, step2, NotAStep())

        self.assertEquals(step1.step_number, 1)
        self.assertEquals(step2.step_number, 2)

    def test_the_number_does_not_represent_the_order_of_creation(self):
        block = Parent()
        step1 = Step()
        step2 = Step()
        block._set_children_for_test(step2, 1, "2", "Step", NotAStep(), False, step1, NotAStep())

        self.assertEquals(step1.step_number, 2)
        self.assertEquals(step2.step_number, 1)

    def test_lonely_child_is_true_for_stand_alone_steps(self):
        block = Parent()
        step1 = Step()
        block._set_children_for_test(1, "2", step1, "Step", NotAStep(), False)

        self.assertTrue(step1.lonely_child)

    def test_lonely_child_is_true_if_parent_have_more_steps(self):
        block = Parent()
        step1 = Step()
        step2 = Step()
        block._set_children_for_test(1, step2, "2", step1, "Step", NotAStep(), False)

        self.assertFalse(step1.lonely_child)
        self.assertFalse(step2.lonely_child)


class TestMentoringStep(BlockWithChildrenTestMixin, unittest.TestCase):

    def test_allowed_nested_blocks(self):
        block = MentoringStepBlock(Mock(), DictFieldData({}), Mock())
        self.assert_allowed_nested_blocks(block)



# -*- coding: utf-8 -*-
#
# Copyright (c) 2014-2015 Harvard, edX & OpenCraft
#
# This software's license gives you freedom; you can copy, convey,
# propagate, redistribute and/or modify this program under the terms of
# the GNU Affero General Public License (AGPL) as published by the Free
# Software Foundation (FSF), either version 3 of the License, or (at your
# option) any later version of the AGPL published by the FSF.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero
# General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program in a file in the toplevel directory called
# "AGPLv3".  If not, see <http://www.gnu.org/licenses/>.
#

import json
import logging

import six
from lazy.lazy import lazy
from xblock.core import XBlock
from xblock.fields import Scope, String
from xblock.fragment import Fragment
from xblock.validation import ValidationMessage
from xblockutils.helpers import child_isinstance
from xblockutils.resources import ResourceLoader
from xblockutils.studio_editable import (StudioContainerWithNestedXBlocksMixin,
                                         StudioEditableXBlockMixin,
                                         XBlockWithPreviewMixin)

from .mixins import StudentViewUserStateMixin
from .sub_api import sub_api

loader = ResourceLoader(__name__)

log = logging.getLogger(__name__)


# Make '_' a no-op so we can scrape strings
def _(text):
    return text


def _normalize_id(key):
    """
    Helper method to normalize a key to avoid issues where some keys have version/branch and others don't.
    e.g. self.scope_ids.usage_id != self.runtime.get_block(self.scope_ids.usage_id).scope_ids.usage_id
    """
    if hasattr(key, "for_branch"):
        key = key.for_branch(None)
    if hasattr(key, "for_version"):
        key = key.for_version(None)
    return key


@XBlock.needs('i18n')
@XBlock.wants('user')
class PlotBlock(
    StudioEditableXBlockMixin, StudioContainerWithNestedXBlocksMixin, XBlockWithPreviewMixin, XBlock,
    StudentViewUserStateMixin,
):
    """
    XBlock that displays plot that summarizes answers to scale and/or rating questions.
    """

    CATEGORY = 'sb-plot'
    STUDIO_LABEL = _(u"Plot")

    # Settings
    display_name = String(
        display_name=_("Plot title"),
        default="Plot",
        scope=Scope.content
    )

    plot_label = String(
        display_name=_("Plot label"),
        help=_("Label for default overlay that shows student's answers to scale questions"),
        default="yours",
        scope=Scope.content
    )

    point_color_default = String(
        display_name=_("Point color (default overlay)"),
        help=_("Point color to use for default overlay"),
        default="green",
        scope=Scope.content
    )

    point_color_average = String(
        display_name=_("Point color (average overlay)"),
        help=_("Point color to use for average overlay"),
        default="blue",
        scope=Scope.content
    )

    q1_label = String(
        display_name=_("Quadrant I"),
        help=_(
            "Label for the first quadrant. "
            "Plot uses counter-clockwise numbering starting in the top right quadrant."
        ),
        default="Q1",
        scope=Scope.content
    )

    q2_label = String(
        display_name=_("Quadrant II"),
        help=_(
            "Label for the second quadrant. "
            "Plot uses counter-clockwise numbering starting in the top right quadrant."
        ),
        default="Q2",
        scope=Scope.content
    )

    q3_label = String(
        display_name=_("Quadrant III"),
        help=_(
            "Label for the third quadrant. "
            "Plot uses counter-clockwise numbering starting in the top right quadrant."
        ),
        default="Q3",
        scope=Scope.content
    )

    q4_label = String(
        display_name=_("Quadrant IV"),
        help=_(
            "Label for the fourth quadrant. "
            "Plot uses counter-clockwise numbering starting in the top right quadrant."
        ),
        default="Q4",
        scope=Scope.content
    )

    claims = String(
        display_name=_("Claims and associated questions"),
        help=_(
            'Claims and questions that should be included in the plot. '
            'Each line defines a triple of the form "claim, q1, q2", '
            'where "claim" is arbitrary text that represents a claim, '
            'and "q1" and "q2" are IDs of scale or rating questions. '
        ),
        default="",
        multiline_editor=True,
        resettable_editor=False
    )

    editable_fields = (
        'display_name', 'plot_label', 'point_color_default', 'point_color_average',
        'q1_label', 'q2_label', 'q3_label', 'q4_label', 'claims'
    )

    @lazy
    def course_key_str(self):
        location = _normalize_id(self.location)
        return six.text_type(location.course_key)

    @property
    def default_claims(self):
        return self._get_claims(self._get_default_response)

    @property
    def average_claims(self):
        return self._get_claims(self._get_average_response)

    def _get_claims(self, response_function):
        if not self.claims:
            return []

        mentoring_block = self.get_parent().get_parent()
        question_ids, questions = mentoring_block.question_ids, mentoring_block.questions
        claims = []
        for line in self.claims.split('\n'):
            claim, q1, q2 = line.split(', ')
            r1, r2 = None, None
            for question_id, question in zip(question_ids, questions):
                if question.url_name == q1:
                    r1 = response_function(question, question_id)
                if question.url_name == q2:
                    r2 = response_function(question, question_id)
                if r1 is not None and r2 is not None:
                    break
            claims.append([claim, r1, r2])

        return claims

    def _get_default_response(self, question, question_id):
        # 1. Obtain block_type for question
        question_type = question.scope_ids.block_type
        # 2. Obtain latest submission for question
        student_dict = {
            'student_id': self.runtime.anonymous_student_id,
            'course_id': self.course_key_str,
            'item_id': question_id,
            'item_type': question_type,
        }
        submissions = sub_api.get_submissions(student_dict, limit=1)
        # 3. Extract response from latest submission for question
        answer_cache = {}
        for submission in submissions:
            answer = self._get_answer(question, submission, answer_cache)
            return int(answer)

    def _get_average_response(self, question, question_id):
        # 1. Obtain block_type for question
        question_type = question.scope_ids.block_type
        # 2. Obtain latest submissions for question
        submissions = sub_api.get_all_submissions(self.course_key_str, question_id, question_type)
        # 3. Extract responses from latest submissions for question and sum them up
        answer_cache = {}
        response_total = 0
        num_submissions = 0  # Can't use len(submissions) because submissions is a generator
        for submission in submissions:
            answer = self._get_answer(question, submission, answer_cache)
            response_total += int(answer)
            num_submissions += 1
        # 4. Calculate average response for question
        if num_submissions:
            return response_total / float(num_submissions)

    def _get_answer(self, block, submission, answer_cache):
        """
        Return answer associated with `submission` to `block`.

        `answer_cache` is a dict that is reset for each block.
        """
        answer = submission['answer']
        # Convert from answer ID to answer label
        if answer not in answer_cache:
            answer_cache[answer] = block.get_submission_display(answer)
        return answer_cache[answer]

    def default_claims_json(self):
        return json.dumps(self.default_claims)

    def average_claims_json(self):
        return json.dumps(self.average_claims)

    def build_user_state_data(self, context=None):
        user_state_data = super(PlotBlock, self).build_user_state_data()
        user_state_data['default_claims'] = self.default_claims
        user_state_data['average_claims'] = self.average_claims
        return user_state_data

    @XBlock.json_handler
    def get_data(self, data, suffix):
        return {
            'default_claims': self.default_claims,
            'average_claims': self.average_claims,
        }

    @property
    def allowed_nested_blocks(self):
        """
        Returns a list of allowed nested XBlocks. Each item can be either

        * An XBlock class
        * A NestedXBlockSpec

        If XBlock class is used it is assumed that this XBlock is enabled and allows multiple instances.
        NestedXBlockSpec allows explicitly setting disabled/enabled state,
        disabled reason (if any) and single/multiple instances.
        """
        return [PlotOverlayBlock]

    @lazy
    def overlay_ids(self):
        """
        Get the usage_ids of all of this XBlock's children that are overlays.
        """
        return [
            _normalize_id(child_id) for child_id in self.children if
            child_isinstance(self, child_id, PlotOverlayBlock)
        ]

    @lazy
    def overlays(self):
        """
        Get the overlay children of this block.
        """
        return [self.runtime.get_block(overlay_id) for overlay_id in self.overlay_ids]

    @lazy
    def overlay_data(self):
        if not self.claims:
            return []

        overlay_data = []
        claims = self.claims.split('\n')
        for index, overlay in enumerate(self.overlays):
            claims_json = []
            if overlay.claim_data:
                claim_data = overlay.claim_data.split('\n')
                for claim, data in zip(claims, claim_data):
                    claim = claim.split(', ')[0]
                    r1, r2 = data.split(', ')
                    claims_json.append([claim, int(r1), int(r2)])
            claims_json = json.dumps(claims_json)
            overlay_data.append({
                'plot_label': overlay.plot_label,
                'point_color': overlay.point_color,
                'description': overlay.description,
                'citation': overlay.citation,
                'claims_json': claims_json,
                'position': index,
            })
        return overlay_data

    @lazy
    def claims_display(self):
        if not self.claims:
            return []

        claims = []
        for claim in self.claims.split('\n'):
            claim, q1, q2 = claim.split(', ')
            claims.append([claim, q1, q2])
        return claims

    def author_preview_view(self, context):
        context['self'] = self
        fragment = Fragment()
        fragment.add_content(loader.render_template('templates/html/plot_preview.html', context))
        fragment.add_css_url(self.runtime.local_resource_url(self, 'public/css/plot-preview.css'))
        if self.overlay_ids:
            fragment.add_content(
                u"<p>{}</p>".format(
                    _(u"In addition to the default and average overlays the plot includes the following overlays:")
                ))
            for overlay in self.overlays:
                overlay_fragment = self._render_child_fragment(overlay, context, view='mentoring_view')
                fragment.add_frag_resources(overlay_fragment)
                fragment.add_content(overlay_fragment.content)
        return fragment

    def mentoring_view(self, context):
        return self.student_view(context)

    def student_view(self, context=None):
        """ Student View """
        context = context.copy() if context else {}
        context['hide_header'] = True
        context['self'] = self
        fragment = Fragment()
        fragment.add_content(loader.render_template('templates/html/plot.html', context))
        fragment.add_css_url(self.runtime.local_resource_url(self, 'public/css/plot.css'))
        fragment.add_javascript_url(self.runtime.local_resource_url(self, 'public/js/vendor/d3.min.js'))
        fragment.add_javascript_url(self.runtime.local_resource_url(self, 'public/js/plot.js'))
        fragment.initialize_js('PlotBlock')
        return fragment

    def student_view_data(self, context=None):
        """
        Returns a JSON representation of the student_view of this XBlock,
        retrievable from the Course XBlock API.
        """
        return {
            'display_name': self.display_name,
            'type': self.CATEGORY,
            'title': self.display_name,
            'q1_label': self.q1_label,
            'q2_label': self.q2_label,
            'q3_label': self.q3_label,
            'q4_label': self.q4_label,
            'point_color_default': self.point_color_default,
            'plot_label': self.plot_label,
            'point_color_average': self.point_color_average,
            'overlay_data': self.overlay_data,
            'hide_header': True,
            'claims': self.claims,
        }

    def author_edit_view(self, context):
        """
        Add some HTML to the author view that allows authors to add child blocks.
        """
        fragment = super(PlotBlock, self).author_edit_view(context)
        fragment.add_css_url(self.runtime.local_resource_url(self, 'public/css/problem-builder-edit.css'))
        fragment.add_javascript_url(self.runtime.local_resource_url(self, 'public/js/util.js'))
        fragment.add_javascript_url(self.runtime.local_resource_url(self, 'public/js/container_edit.js'))
        fragment.initialize_js('ProblemBuilderContainerEdit')
        return fragment


@XBlock.needs('i18n')
class PlotOverlayBlock(StudioEditableXBlockMixin, XBlockWithPreviewMixin, XBlock):
    """
    XBlock that represents a user-defined overlay for a plot block.
    """

    CATEGORY = 'sb-plot-overlay'
    STUDIO_LABEL = _(u"Plot Overlay")

    # Settings
    display_name = String(
        display_name=_("Overlay title"),
        default="Overlay",
        scope=Scope.content
    )

    plot_label = String(
        display_name=_("Plot label"),
        help=_("Label for button that allows to toggle visibility of this overlay"),
        default="",
        scope=Scope.content
    )

    point_color = String(
        display_name=_("Point color"),
        help=_("Point color to use for this overlay"),
        default="",
        scope=Scope.content
    )

    description = String(
        display_name=_("Description"),
        help=_("Description of this overlay (optional)"),
        default="",
        scope=Scope.content
    )

    citation = String(
        display_name=_("Citation"),
        help=_("Source of data belonging to this overlay (optional)"),
        default="",
        scope=Scope.content
    )

    claim_data = String(
        display_name=_("Claim data"),
        help=_(
            'Claim data to include in this overlay. '
            'Each line defines a tuple of the form "q1, q2", '
            'where "q1" is the value associated with the first scale or rating question, '
            'and "q2" is the value associated with the second scale or rating question. '
            'Note that data will be associated with claims in the order that they are defined in the parent plot.'
        ),
        default="",
        multiline_editor=True,
        resettable_editor=False
    )

    editable_fields = (
        "plot_label", "point_color", "description", "citation", "claim_data"
    )

    def validate_field_data(self, validation, data):
        """
        Validate this block's field data.
        """
        super(PlotOverlayBlock, self).validate_field_data(validation, data)

        def add_error(msg):
            validation.add(ValidationMessage(ValidationMessage.ERROR, msg))

        if not data.plot_label.strip():
            add_error(_(u"No plot label set. Button for toggling visibility of this overlay will not have a label."))
        if not data.point_color.strip():
            add_error(_(u"No point color set. This overlay will not work correctly."))

        # If parent plot is associated with one or more claims, prompt user to add claim data
        parent = self.get_parent()
        if parent.claims.strip() and not data.claim_data.strip():
            add_error(_(u"No claim data provided. This overlay will not work correctly."))

    def author_preview_view(self, context):
        return self.student_view(context)

    def mentoring_view(self, context):
        context = context.copy() if context else {}
        context['hide_header'] = True
        return self.author_preview_view(context)

    def student_view(self, context):
        context['self'] = self
        fragment = Fragment()
        fragment.add_content(loader.render_template('templates/html/overlay.html', context))
        fragment.add_css_url(self.runtime.local_resource_url(self, 'public/css/overlay.css'))
        return fragment



# -*- coding: utf-8 -*-
#
# Copyright (c) 2014-2015 Harvard, edX & OpenCraft
#
# This software's license gives you freedom; you can copy, convey,
# propagate, redistribute and/or modify this program under the terms of
# the GNU Affero General Public License (AGPL) as published by the Free
# Software Foundation (FSF), either version 3 of the License, or (at your
# option) any later version of the AGPL published by the FSF.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero
# General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program in a file in the toplevel directory called
# "AGPLv3".  If not, see <http://www.gnu.org/licenses/>.
#

# Imports ###########################################################
import logging
import uuid

import pkg_resources
import six
from lazy import lazy
from django import utils
from xblock.core import XBlock
from xblock.fields import Integer, Scope, String
from xblock.fragment import Fragment
from xblock.validation import ValidationMessage
from xblockutils.resources import ResourceLoader
from xblockutils.studio_editable import (StudioEditableXBlockMixin,
                                         XBlockWithPreviewMixin)

from problem_builder.sub_api import SubmittingXBlockMixin, sub_api

from .mixins import (ExpandStaticURLMixin, QuestionMixin,
                     StudentViewUserStateMixin,
                     XBlockWithTranslationServiceMixin)
from .models import Answer
from .utils import I18NService

# Globals ###########################################################

log = logging.getLogger(__name__)
loader = ResourceLoader(__name__)


# Make '_' a no-op so we can scrape strings
def _(text):
    return text

# Classes ###########################################################


class AnswerMixin(XBlockWithPreviewMixin, XBlockWithTranslationServiceMixin, StudentViewUserStateMixin, I18NService):
    """
    Mixin to give an XBlock the ability to read/write data to the Answers DB table.
    """
    def build_user_state_data(self, context=None):
        result = super(AnswerMixin, self).build_user_state_data()
        result['student_input'] = self.student_input
        return result

    def _get_course_id(self):
        """ Get a course ID if available """
        return getattr(self.runtime, 'course_id', 'all')

    def _get_student_id(self):
        """ Get student anonymous ID or normal ID """
        try:
            return self.runtime.anonymous_student_id
        except AttributeError:
            return self.scope_ids.user_id

    def get_model_object(self, name=None):
        """
        Fetches the Answer model object for the answer named `name`
        """
        # By default, get the model object for the current answer's name
        if not name:
            name = self.name
        # Consistency check - we should have a name by now
        if not name:
            raise ValueError('AnswerBlock.name field need to be set to a non-null/empty value')

        student_id = self._get_student_id()
        course_key = self._get_course_id()

        answer_data, _ = Answer.objects.get_or_create(
            student_id=student_id,
            course_key=course_key,
            name=name,
        )

        return answer_data

    @XBlock.json_handler
    def answer_value(self, data, suffix=''):
        """ Current value of the answer, for refresh by client """
        return {'value': self.student_input}

    @XBlock.json_handler
    def refresh_html(self, data, suffix=''):
        """ Complete HTML view of the XBlock, for refresh by client """
        frag = self.mentoring_view({})
        return {'html': frag.content}

    @lazy
    def student_input(self):
        """
        The student input value, or a default which may come from another block.
        Read from a Django model, since XBlock API doesn't yet support course-scoped
        fields or generating instructor reports across many student responses.
        """
        # Only attempt to locate a model object for this block when the answer has a name
        if not self.name:
            return ''

        student_input = self.get_model_object().student_input

        # Default value can be set from another answer's current value
        if not student_input and hasattr(self, 'default_from') and self.default_from:
            student_input = self.get_model_object(name=self.default_from).student_input

        return student_input

    def validate_field_data(self, validation, data):
        """
        Validate this block's field data.
        """
        super(AnswerMixin, self).validate_field_data(validation, data)

        def add_error(msg):
            validation.add(ValidationMessage(ValidationMessage.ERROR, msg))

        if not data.name:
            add_error(u"A Question ID is required.")


@XBlock.needs("i18n")
class AnswerBlock(SubmittingXBlockMixin, AnswerMixin, QuestionMixin, StudioEditableXBlockMixin,
                  XBlock, ExpandStaticURLMixin):
    """
    A field where the student enters an answer

    Must be included as a child of a mentoring block. Answers are persisted as django model instances
    to make them searchable and referenceable across xblocks.
    """
    CATEGORY = 'pb-answer'
    STUDIO_LABEL = _(u"Long Answer")
    answerable = True

    name = String(
        display_name=_("Question ID (name)"),
        help=_("The ID of this block. Should be unique unless you want the answer to be used in multiple places."),
        default="",
        scope=Scope.content
    )
    default_from = String(
        display_name=_("Default From"),
        help=_("If a question ID is specified, get the default value from this answer."),
        default=None,
        scope=Scope.content
    )
    min_characters = Integer(
        display_name=_("Min. Allowed Characters"),
        help=_("Minimum number of characters allowed for the answer"),
        default=0,
        scope=Scope.content
    )
    question = String(
        display_name=_("Question"),
        help=_("Question to ask the student"),
        scope=Scope.content,
        default="",
        multiline_editor=True,
    )

    editable_fields = ('question', 'name', 'min_characters', 'weight', 'default_from', 'display_name', 'show_title')

    @staticmethod
    def resource_string(path):
        """Handy helper for getting resources from our kit."""
        data = pkg_resources.resource_string(__name__, path)
        return data.decode("utf8")

    def get_translation_content(self):
        try:
            return self.resource_string('public/js/translations/{lang}/textjs.js'.format(
                lang=utils.translation.to_locale(utils.translation.get_language()),
            ))
        except IOError:
            return self.resource_string('public/js/translations/en/textjs.js')

    def mentoring_view(self, context=None):
        """ Render this XBlock within a mentoring block. """
        context = context.copy() if context else {}
        context['answer_editable_id'] = uuid.uuid4().hex[:15]
        context['self'] = self
        context['hide_header'] = context.get('hide_header', False) or not self.show_title
        html = loader.render_django_template(
            'templates/html/answer_editable.html',
            context,
            i18n_service=self.i18n_service
        )

        fragment = Fragment(html)
        fragment.add_css_url(self.runtime.local_resource_url(self, 'public/css/answer.css'))
        fragment.add_javascript_url(self.runtime.local_resource_url(self, 'public/js/answer.js'))
        fragment.add_javascript(self.get_translation_content())
        fragment.initialize_js('AnswerBlock')
        return fragment

    def student_view(self, context=None):
        """ Normal view of this XBlock, identical to mentoring_view """
        return self.mentoring_view(context)

    def get_results(self, previous_response=None):
        # Previous result is actually stored in database table-- ignore.
        return {
            'student_input': self.student_input,
            'status': self.status,
            'weight': self.weight,
            'score': 1 if self.status == 'correct' else 0,
        }

    def get_last_result(self):
        return self.get_results(None) if self.student_input else {}

    def submit(self, submission):
        """
        The parent block is handling a student submission, including a new answer for this
        block. Update accordingly.
        """
        self.student_input = submission['value'].strip()
        self.save()

        if sub_api:
            # Also send to the submissions API:
            item_key = self.student_item_key
            # Need to do this by our own ID, since an answer can be referred to multiple times.
            item_key['item_id'] = self.name
            sub_api.create_submission(item_key, self.student_input)

        log.info(u'Answer submitted for`{}`: "{}"'.format(self.name, self.student_input))
        return self.get_results()

    @property
    def status(self):
        answer_length_ok = self.student_input
        if self.min_characters and self.min_characters > 0:
            answer_length_ok = len(self.student_input.strip()) >= self.min_characters

        return 'correct' if answer_length_ok else 'incorrect'

    @property
    def completed(self):
        return self.status == 'correct'

    def save(self):
        """
        Replicate data changes on the related Django model used for sharing of data accross XBlocks
        """
        super(AnswerBlock, self).save()

        student_id = self._get_student_id()
        if not student_id:
            return  # save() gets called from the workbench homepage sometimes when there is no student ID

        # Only attempt to locate a model object for this block when the answer has a name
        if self.name:
            answer_data = self.get_model_object()
            if answer_data.student_input != self.student_input:
                answer_data.student_input = self.student_input
                answer_data.save()

    @classmethod
    def get_template(cls, template_id):
        """
        Used to interact with Studio's create_xblock method to instantiate pre-defined templates.
        """
        # Generate a random 'name' value
        if template_id == 'studio_default':
            return {'data': {'name': uuid.uuid4().hex[:7]}}
        return {'metadata': {}, 'data': {}}

    def student_view_data(self, context=None):
        """
        Returns a JSON representation of the student_view of this XBlock,
        retrievable from the Course Block API.
        """
        return {
            'id': self.name,
            'block_id': six.text_type(self.scope_ids.usage_id),
            'display_name': self.display_name,
            'type': self.CATEGORY,
            'weight': self.weight,
            'question': self.expand_static_url(self.question),
            'name': self.name,  # For backwards compatibility; same as 'id'
        }


@XBlock.needs("i18n")
class AnswerRecapBlock(AnswerMixin, StudioEditableXBlockMixin, XBlock):
    """
    A block that displays an answer previously entered by the student (read-only).
    """

    CATEGORY = 'pb-answer-recap'
    STUDIO_LABEL = _(u"Long Answer Recap")

    name = String(
        display_name=_("Question ID"),
        help=_("The ID of the question for which to display the student's answer."),
        scope=Scope.content,
    )
    display_name = String(
        display_name=_("Title"),
        help=_("Title of this answer recap section"),
        scope=Scope.content,
        default="",
    )
    description = String(
        display_name=_("Description"),
        help=_("Description of this answer (optional). Can include HTML."),
        scope=Scope.content,
        default="",
    )
    editable_fields = ('name', 'display_name', 'description')

    css_path = 'public/css/answer.css'

    def mentoring_view(self, context=None):
        """ Render this XBlock within a mentoring block. """
        context = context.copy() if context else {}
        student_submissions_key = context.get('student_submissions_key')
        context['title'] = self.display_name
        context['description'] = self.description
        if student_submissions_key:
            location = self.location.replace(branch=None, version=None)  # Standardize the key in case it isn't already
            target_key = {
                'student_id': student_submissions_key,
                'course_id': six.text_type(location.course_key),
                'item_id': self.name,
                'item_type': u'pb-answer',
            }
            submissions = sub_api.get_submissions(target_key, limit=1)
            try:
                context['student_input'] = submissions[0]['answer']
            except IndexError:
                context['student_input'] = None
        else:
            context['student_input'] = self.student_input
        html = loader.render_django_template(
            'templates/html/answer_read_only.html',
            context,
            i18n_service=self.i18n_service
        )

        fragment = Fragment(html)
        fragment.add_css_url(self.runtime.local_resource_url(self, self.css_path))
        fragment.add_javascript_url(self.runtime.local_resource_url(self, 'public/js/answer_recap.js'))
        fragment.initialize_js('AnswerRecapBlock')
        return fragment

    def student_view(self, context=None):
        """ Normal view of this XBlock, identical to mentoring_view """
        return self.mentoring_view(context)

    def student_view_data(self, context=None):
        """
        Returns a JSON representation of the student_view of this XBlock,
        retrievable from the Course Block API.
        """

        return {
            'id': self.name,
            'name': self.name,  # For backwards compatibility; same as 'id'
            'display_name': self.display_name,
            'description': self.description,
            'block_id': six.text_type(self.scope_ids.usage_id),
            'type': self.CATEGORY
        }



# -*- coding: utf-8 -*-
#
# Copyright (c) 2014-2015 Harvard, edX & OpenCraft
#
# This software's license gives you freedom; you can copy, convey,
# propagate, redistribute and/or modify this program under the terms of
# the GNU Affero General Public License (AGPL) as published by the Free
# Software Foundation (FSF), either version 3 of the License, or (at your
# option) any later version of the AGPL published by the FSF.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero
# General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program in a file in the toplevel directory called
# "AGPLv3".  If not, see <http://www.gnu.org/licenses/>.
#

# Imports ###########################################################

import pkg_resources
import uuid

from django import utils
from django.utils.safestring import mark_safe
from lazy import lazy
from xblock.core import XBlock
from xblock.fields import Scope, String
from xblock.fragment import Fragment
from xblock.validation import ValidationMessage
from xblockutils.helpers import child_isinstance
from xblockutils.resources import ResourceLoader
from xblockutils.studio_editable import (StudioContainerXBlockMixin,
                                         StudioEditableXBlockMixin,
                                         XBlockWithPreviewMixin)

from .choice import ChoiceBlock
from .message import MentoringMessageBlock
from .mixins import QuestionMixin, XBlockWithTranslationServiceMixin
from .tip import TipBlock
from .utils import I18NService

# Globals ###########################################################

loader = ResourceLoader(__name__)


# Make '_' a no-op so we can scrape strings
def _(text):
    return text

# Classes ###########################################################


@XBlock.needs("i18n")
class QuestionnaireAbstractBlock(
    StudioEditableXBlockMixin, StudioContainerXBlockMixin, QuestionMixin, XBlock, XBlockWithPreviewMixin,
    XBlockWithTranslationServiceMixin, I18NService
):
    """
    An abstract class used for MCQ/MRQ blocks

    Must be a child of a MentoringBlock. Allow to display a tip/advice depending on the
    values entered by the student, and supports multiple types of multiple-choice
    set, with preset choices and author-defined values.
    """
    question = String(
        display_name=_("Question"),
        help=_("Question to ask the student"),
        scope=Scope.content,
        default="",
        multiline_editor=True,
    )

    editable_fields = ('question', 'weight', 'display_name', 'show_title')
    has_children = True
    answerable = True

    @lazy
    def html_id(self):
        """
        A short, simple ID string used to uniquely identify this question.

        This is only used by templates for matching <input> and <label> elements.
        """
        return uuid.uuid4().hex[:20]

    @staticmethod
    def resource_string(path):
        """Handy helper for getting resources from our kit."""
        data = pkg_resources.resource_string(__name__, path)
        return data.decode("utf8")

    def get_translation_content(self):
        try:
            return self.resource_string('public/js/translations/{lang}/textjs.js'.format(
                lang=utils.translation.to_locale(utils.translation.get_language()),
            ))
        except IOError:
            return self.resource_string('public/js/translations/en/textjs.js')

    def student_view(self, context=None):
        name = getattr(self, "unmixed_class", self.__class__).__name__

        template_path = 'templates/html/{}.html'.format(name.lower())

        context = context.copy() if context else {}
        context['self'] = self
        context['custom_choices'] = self.custom_choices
        context['hide_header'] = context.get('hide_header', False) or not self.show_title

        fragment = Fragment(loader.render_django_template(template_path, context, i18n_service=self.i18n_service))
        # If we use local_resource_url(self, ...) the runtime may insert many identical copies
        # of questionnaire.[css/js] into the DOM. So we use the mentoring block here if possible.
        block_with_resources = self.get_parent()
        from .mentoring import MentoringBlock
        # We use an inline import here to avoid a circular dependency with the .mentoring module.
        if not isinstance(block_with_resources, MentoringBlock):
            block_with_resources = self
        fragment.add_css_url(self.runtime.local_resource_url(block_with_resources, 'public/css/questionnaire.css'))
        fragment.add_javascript_url(self.runtime.local_resource_url(block_with_resources, 'public/js/questionnaire.js'))
        fragment.add_javascript(self.get_translation_content())
        fragment.initialize_js(name)
        return fragment

    def mentoring_view(self, context=None):
        return self.student_view(context)

    @property
    def custom_choices(self):
        custom_choices = []
        for child_id in self.children:
            if child_isinstance(self, child_id, ChoiceBlock):
                custom_choices.append(self.runtime.get_block(child_id))
        return custom_choices

    @property
    def all_choice_values(self):
        return [c.value for c in self.custom_choices]

    @property
    def human_readable_choices(self):
        return [{"display_name": mark_safe(c.content), "value": c.value} for c in self.custom_choices]

    @staticmethod
    def choice_values_provider(question):
        """
        Get a list a {"display_name": "Choice Description", "value": value}
        objects for use with studio_view editor.
        """
        return question.human_readable_choices

    def get_tips(self):
        """
        Returns the tips contained in this block
        """
        tips = []
        for child_id in self.children:
            if child_isinstance(self, child_id, TipBlock):
                tips.append(self.runtime.get_block(child_id))
        return tips

    def get_submission_display(self, submission):
        """
        Get the human-readable version of a submission value
        """
        for choice in self.custom_choices:
            if choice.value == submission:
                return choice.content
        return submission

    def get_author_edit_view_fragment(self, context):
        fragment = super(QuestionnaireAbstractBlock, self).author_edit_view(context)
        return fragment

    def author_edit_view(self, context):
        """
        Add some HTML to the author view that allows authors to add choices and tips.
        """
        from .mentoring import MentoringWithExplicitStepsBlock

        fragment = self.get_author_edit_view_fragment(context)

        # * Step Builder units can show review components in the Review Step.
        show_review = isinstance(
            self.get_parent().get_parent() if self.get_parent() else None,
            MentoringWithExplicitStepsBlock
        )
        fragment.add_content(loader.render_template('templates/html/questionnaire_add_buttons.html', {
            'show_review': show_review
        }))
        fragment.add_css_url(self.runtime.local_resource_url(self, 'public/css/problem-builder.css'))
        fragment.add_css_url(self.runtime.local_resource_url(self, 'public/css/questionnaire-edit.css'))
        fragment.add_javascript_url(self.runtime.local_resource_url(self, 'public/js/util.js'))
        fragment.add_javascript_url(self.runtime.local_resource_url(self, 'public/js/mentoring_edit.js'))
        fragment.initialize_js('MentoringEditComponents')
        return fragment

    def validate_field_data(self, validation, data):
        """
        Validate this block's field data. Instead of checking fields like self.name, check the
        fields set on data, e.g. data.name. This allows the same validation method to be re-used
        for the studio editor. Any errors found should be added to "validation".

        This method should not return any value or raise any exceptions.
        All of this XBlock's fields should be found in "data", even if they aren't being changed
        or aren't even set (i.e. are defaults).
        """
        super(QuestionnaireAbstractBlock, self).validate_field_data(validation, data)

        def add_error(msg):
            validation.add(ValidationMessage(ValidationMessage.ERROR, msg))
        if not data.name:
            add_error(self._(u"A unique Question ID is required."))
        elif ' ' in data.name:
            add_error(self._(u"Question ID should not contain spaces."))

    def validate(self):
        """
        Validates the state of this XBlock.
        """
        validation = super(QuestionnaireAbstractBlock, self).validate()

        def add_error(msg):
            validation.add(ValidationMessage(ValidationMessage.ERROR, msg))

        # Validate the choice values:
        all_choice_values = self.all_choice_values
        all_choice_values_set = set(all_choice_values)
        if len(all_choice_values) != len(all_choice_values_set):
            add_error(self._(u"Some choice values are not unique."))
        # Validate the tips:
        values_with_tips = set()
        for tip in self.get_tips():
            values = set(tip.values)
            if values & values_with_tips:
                add_error(self._(u"Multiple tips configured for the same choice."))
                break
            values_with_tips.update(values)
        return validation

    def get_review_tip(self):
        """ Get the text to show on the assessment review when the student gets this question wrong """
        for child_id in self.children:
            if child_isinstance(self, child_id, MentoringMessageBlock):
                child = self.runtime.get_block(child_id)
                if child.type == "on-assessment-review-question":
                    return child.content

    @property
    def message_formatted(self):
        """ Get the feedback message HTML, if any, formatted by the runtime """
        if self.message:
            # For any HTML that we aren't 'rendering' through an XBlock view such as
            # student_view the runtime may need to rewrite URLs
            # e.g. converting '/static/x.png' to '/c4x/.../x.png'
            format_html = getattr(self.runtime, 'replace_urls', lambda html: html)
            return format_html(self.message)
        return ""



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('problem_builder', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='Share',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('submission_uid', models.CharField(max_length=32)),
                ('block_id', models.CharField(max_length=255, db_index=True)),
                ('notified', models.BooleanField(default=False, db_index=True)),
                ('shared_by', models.ForeignKey(related_name='problem_builder_shared_by', to=settings.AUTH_USER_MODEL)),
                ('shared_with', models.ForeignKey(related_name='problem_builder_shared_with', to=settings.AUTH_USER_MODEL)),
            ],
        ),
        migrations.AlterUniqueTogether(
            name='share',
            unique_together=set([('shared_by', 'shared_with', 'block_id')]),
        ),
    ]



# -*- coding: utf-8 -*-
#
# Copyright (c) 2014-2015 Harvard, edX & OpenCraft
#
# This software's license gives you freedom; you can copy, convey,
# propagate, redistribute and/or modify this program under the terms of
# the GNU Affero General Public License (AGPL) as published by the Free
# Software Foundation (FSF), either version 3 of the License, or (at your
# option) any later version of the AGPL published by the FSF.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero
# General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program in a file in the toplevel directory called
# "AGPLv3".  If not, see <http://www.gnu.org/licenses/>.
#
"""
Test that we can upgrade from mentoring v1 to problem builder (v2).
"""
import os.path
import unittest

import ddt
from lxml import etree
from six import StringIO
from xblock.core import XBlock
from xblock.fields import ScopeIds
from xblock.runtime import DictKeyValueStore, KvsFieldData
from xblock.test.tools import TestRuntime

from problem_builder.mentoring import MentoringBlock
from problem_builder.v1.xml_changes import convert_xml_to_v2
from sample_xblocks.basic.content import HtmlBlock

xml_path = os.path.join(os.path.dirname(__file__), "xml")


@ddt.ddt
class TestUpgrade(unittest.TestCase):
    """
    Test upgrade from mentoring v1 (which uses xml_content even in Studio) to v2.

    We can't test the steps that depend on Studio, so we just test the XML conversion.
    """

    def setUp(self):
        self.runtime = TestRuntime(field_data=KvsFieldData(DictKeyValueStore()))

    @ddt.data(
        "v1_upgrade_a",
        "v1_upgrade_b",
        "v1_upgrade_c",
    )
    @XBlock.register_temp_plugin(HtmlBlock, "html")
    @XBlock.register_temp_plugin(MentoringBlock, "mentoring")
    def test_xml_upgrade(self, file_name):
        """
        Convert a v1 mentoring block to v2 and then compare the resulting block to a pre-converted one.
        """
        with open("{}/{}_old.xml".format(xml_path, file_name)) as xmlfile:
            temp_node = etree.parse(xmlfile).getroot()
            old_block = self.create_block_from_node(temp_node)

        parser = etree.XMLParser(remove_blank_text=True)
        xml_root = etree.parse(StringIO(old_block.xml_content), parser=parser).getroot()
        convert_xml_to_v2(xml_root)
        converted_block = self.create_block_from_node(xml_root)

        with open("{}/{}_new.xml".format(xml_path, file_name)) as xmlfile:
            temp_node = etree.parse(xmlfile).getroot()
            new_block = self.create_block_from_node(temp_node)

        try:
            self.assertBlocksAreEquivalent(converted_block, new_block)
        except AssertionError:
            xml_result = etree.tostring(xml_root, pretty_print=True, encoding="UTF-8")
            print("Converted XML:\n{}".format(xml_result))
            raise

    def create_block_from_node(self, node):
        """
        Parse an XML node representing an XBlock (and children), and return the XBlock.
        """
        block_type = node.tag
        def_id = self.runtime.id_generator.create_definition(block_type)
        usage_id = self.runtime.id_generator.create_usage(def_id)
        keys = ScopeIds(None, block_type, def_id, usage_id)
        block_class = self.runtime.mixologist.mix(self.runtime.load_block_type(block_type))
        block = block_class.parse_xml(node, self.runtime, keys, self.runtime.id_generator)
        block.save()
        return block

    def assertBlocksAreEquivalent(self, block1, block2):
        """
        Compare two blocks for equivalence.
        Borrowed from xblock.test.tools.blocks_are_equivalent but modified to use assertions.
        """
        # The two blocks have to be the same class.
        self.assertEqual(block1.__class__, block2.__class__)
        # They have to have the same fields.
        self.assertEqual(set(block1.fields), set(block2.fields))
        # The data fields have to have the same values.
        for field_name in block1.fields:
            if field_name in ('parent', 'children'):
                continue
            if field_name == "content":
                # Inner HTML/XML content may have varying whitespace which we don't care about:
                self.assertEqual(
                    self.clean_html(getattr(block1, field_name)),
                    self.clean_html(getattr(block2, field_name))
                )
            else:
                self.assertEqual(getattr(block1, field_name), getattr(block2, field_name))
        # The children need to be equal.
        self.assertEqual(block1.has_children, block2.has_children)

        if block1.has_children:
            self.assertEqual(len(block1.children), len(block2.children))
            for child_id1, child_id2 in zip(block1.children, block2.children):
                # Load up the actual children to see if they are equal.
                child1 = block1.runtime.get_block(child_id1)
                child2 = block2.runtime.get_block(child_id2)
                self.assertBlocksAreEquivalent(child1, child2)

    def clean_html(self, html_str):
        """
        Standardize the given HTML string for a consistent comparison.
        Assumes the HTML is valid XML.
        """
        # We wrap it in <x></x> so that the given HTML string doesn't need a single root element.
        parser = etree.XMLParser(remove_blank_text=True)
        parsed = etree.parse(StringIO(u"<x>{}</x>".format(html_str)), parser=parser).getroot()
        return etree.tostring(parsed, pretty_print=False, encoding="UTF-8")[3:-3]



# -*- coding: utf-8 -*-
#
# Copyright (c) 2014-2015 Harvard, edX & OpenCraft
#
# This software's license gives you freedom; you can copy, convey,
# propagate, redistribute and/or modify this program under the terms of
# the GNU Affero General Public License (AGPL) as published by the Free
# Software Foundation (FSF), either version 3 of the License, or (at your
# option) any later version of the AGPL published by the FSF.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero
# General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program in a file in the toplevel directory called
# "AGPLv3".  If not, see <http://www.gnu.org/licenses/>.
#

# Imports ###########################################################

import logging

import six
from xblock.core import XBlock
from xblock.fields import Boolean, Scope, String
from xblock.fragment import Fragment
from xblockutils.resources import ResourceLoader
from xblockutils.studio_editable import (StudioEditableXBlockMixin,
                                         XBlockWithPreviewMixin)

from .mixins import QuestionMixin, StudentViewUserStateMixin

# Globals ###########################################################

log = logging.getLogger(__name__)
loader = ResourceLoader(__name__)


# Make '_' a no-op so we can scrape strings
def _(text):
    return text


# Classes ###########################################################

@XBlock.needs("i18n")
class SwipeBlock(
    QuestionMixin, StudioEditableXBlockMixin, StudentViewUserStateMixin, XBlockWithPreviewMixin, XBlock,
):
    """
    An XBlock used to ask binary-choice questions with a swiping interface
    """
    CATEGORY = 'pb-swipe'
    STUDIO_LABEL = _(u"Swipeable Binary Choice Question")
    USER_STATE_FIELDS = ['student_choice']

    text = String(
        display_name=_("Text"),
        help=_("Text to display on this card. The student must determine if this statement is true or false."),
        scope=Scope.content,
        default="",
        multiline_editor=True,
    )

    img_url = String(
        display_name=_("Image"),
        help=_("Specify the URL of an image associated with this question."),
        scope=Scope.content,
        default=""
    )

    correct = Boolean(
        display_name=_("Correct Choice"),
        help=_("Specifies whether the card is correct."),
        scope=Scope.content,
        default=False,
    )

    feedback_correct = String(
        display_name=_("Correct Answer Feedback"),
        help=_("Feedback to display when student answers correctly."),
        scope=Scope.content,
    )

    feedback_incorrect = String(
        display_name=_("Incorrect Answer Feedback"),
        help=_("Feedback to display when student answers incorrectly."),
        scope=Scope.content,
    )

    student_choice = Boolean(
        scope=Scope.user_state,
        help=_("Last input submitted by the student.")
    )

    editable_fields = ('display_name', 'text', 'img_url', 'correct', 'feedback_correct', 'feedback_incorrect')

    def calculate_results(self, submission):
        correct = submission == self.correct
        return {
            'submission': submission,
            'status': 'correct' if correct else 'incorrect',
            'score': 1 if correct else 0,
            'feedback': self.feedback_correct if correct else self.feedback_incorrect,
        }

    def get_results(self, previous_result):
        return self.calculate_results(previous_result['submission'])

    def get_last_result(self):
        return self.get_results({'submission': self.student_choice}) if self.student_choice else {}

    def submit(self, submission):
        log.debug(u'Received Swipe submission: "%s"', submission)
        # We expect to receive a boolean indicating the swipe the student made (left -> false, right -> true)
        self.student_choice = submission['value']
        result = self.calculate_results(self.student_choice)
        log.debug(u'Swipe submission result: %s', result)
        return result

    def student_view_data(self, context=None):
        """
        Returns a JSON representation of the student_view of this XBlock,
        retrievable from the Course Block API.
        """
        return {
            'id': self.name,
            'block_id': six.text_type(self.scope_ids.usage_id),
            'display_name': self.display_name_with_default,
            'type': self.CATEGORY,
            'text': self.text,
            'img_url': self.expand_static_url(self.img_url),
            'correct': self.correct,
            'feedback': {
                'correct': self.feedback_correct,
                'incorrect': self.feedback_incorrect,
            },
        }

    def expand_static_url(self, url):
        """
        This is required to make URLs like '/static/dnd-test-image.png' work (note: that is the
        only portable URL format for static files that works across export/import and reruns).
        This method is unfortunately a bit hackish since XBlock does not provide a low-level API
        for this.
        """
        if hasattr(self.runtime, 'replace_urls'):
            url = self.runtime.replace_urls('"{}"'.format(url))[1:-1]
        elif hasattr(self.runtime, 'course_id'):
            # edX Studio uses a different runtime for 'studio_view' than 'student_view',
            # and the 'studio_view' runtime doesn't provide the replace_urls API.
            try:
                from static_replace import replace_static_urls  # pylint: disable=import-error
                url = replace_static_urls('"{}"'.format(url), None, course_id=self.runtime.course_id)[1:-1]
            except ImportError:
                pass
        return url

    def mentoring_view(self, context=None):
        """ Render the swipe image, text & whether it's correct within a mentoring block question. """
        return Fragment(
            (
                u'<img src="{img_url}" style="max-width: 100%;" />'
                u'<p class="swipe-text">"{text}"</p>'
            ).format(
                img_url=self.expand_static_url(self.img_url),
                text=self.text,
            )
        )

    def student_view(self, context=None):
        """ Normal view of this XBlock, identical to mentoring_view """
        return self.mentoring_view(context)



# -*- coding: utf-8 -*-
#
# Copyright (c) 2014-2015 Harvard, edX & OpenCraft
#
# This software's license gives you freedom; you can copy, convey,
# propagate, redistribute and/or modify this program under the terms of
# the GNU Affero General Public License (AGPL) as published by the Free
# Software Foundation (FSF), either version 3 of the License, or (at your
# option) any later version of the AGPL published by the FSF.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero
# General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program in a file in the toplevel directory called
# "AGPLv3".  If not, see <http://www.gnu.org/licenses/>.
#

# Imports ###########################################################

import errno
import json

import six
from django.contrib.auth.models import User
from xblock.core import XBlock
from xblock.exceptions import JsonHandlerError
from xblock.fields import Boolean, Scope, String
from xblock.fragment import Fragment
from xblockutils.resources import ResourceLoader
from xblockutils.studio_editable import (StudioContainerXBlockMixin,
                                         StudioEditableXBlockMixin,
                                         XBlockWithPreviewMixin)

from problem_builder.answer import AnswerRecapBlock
from problem_builder.dashboard import ExportMixin
from problem_builder.models import Share
from problem_builder.sub_api import SubmittingXBlockMixin

# Globals ###########################################################


loader = ResourceLoader(__name__)


# Make '_' a no-op so we can scrape strings
def _(text):
    return text

# Classes ###########################################################


@XBlock.wants("user")
@XBlock.wants("submissions")
class MentoringTableBlock(
    StudioEditableXBlockMixin, SubmittingXBlockMixin, StudioContainerXBlockMixin, ExportMixin, XBlock,
    XBlockWithPreviewMixin
):
    """
    Table-type display of information from mentoring blocks

    Used to present summary of information entered by the students in mentoring blocks.
    Supports different types of formatting through the `type` parameter.
    """
    CATEGORY = 'pb-table'
    STUDIO_LABEL = _(u"Answer Recap Table")

    display_name = String(
        display_name=_("Display name"),
        help=_("Title of the table"),
        default=_("Answers Table"),
        scope=Scope.settings
    )
    type = String(
        display_name=_("Special Mode"),
        help=_("Variant of the table that will display a specific background image."),
        scope=Scope.content,
        default='',
        values=[
            {"display_name": "Normal", "value": ""},
            {"display_name": "Immunity Map Assumptions", "value": "immunity-map-assumptions"},
            {"display_name": "Immunity Map", "value": "immunity-map"},
        ],
    )
    editable_fields = ("type", "allow_download")
    allow_download = Boolean(
        display_name=_("Allow Download"),
        help=_("Allow students to download a copy of the table for themselves."),
        default=False,
        scope=Scope.content
    )
    allow_sharing = Boolean(
        display_name=_("Allow Sharing"),
        help=_("Allow students to share their results with other students."),
        default=True,
        scope=Scope.content
    )
    has_children = True

    css_path = 'public/css/mentoring-table.css'
    js_path = 'public/js/review_blocks.js'

    @XBlock.json_handler
    def table_render(self, data, suffix=''):
        context = {}
        header_values = []
        content_values = []
        target_username = data.get('target_username')
        try:
            if target_username and target_username != self.current_user_key:
                share = Share.objects.get(
                    shared_by__username=target_username, shared_with__username=self.current_user_key,
                    block_id=self.block_id,
                )
                context['student_submissions_key'] = share.submission_uid
        except Share.DoesNotExist:
            raise JsonHandlerError(403, _("You are not permitted to view this student's table."))

        for child_id in self.children:
            child = self.runtime.get_block(child_id)
            # Child should be an instance of MentoringTableColumn
            header = child.header
            # Make sure /jump_to_id/ URLs are expanded correctly
            if getattr(self.runtime, 'replace_jump_to_id_urls', None):
                header = self.runtime.replace_jump_to_id_urls(header)
            header_values.append(header)
            child_frag = child.render('mentoring_view', context)
            content_values.append(child_frag.content)
        context['header_values'] = header_values if any(header_values) else None
        context['content_values'] = content_values
        html = loader.render_template('templates/html/mentoring-table.html', context)
        return {'content': html}

    @property
    def current_user_key(self):
        user = self.runtime.service(self, 'user').get_current_user()
        # We may be in the SDK, in which case the username may not really be available.
        return user.opt_attrs.get('edx-platform.username', 'username')

    @XBlock.json_handler
    def get_shared_list(self, data, suffix=''):
        context = {'shared_with': Share.objects.filter(
            shared_by__username=self.current_user_key,
            block_id=self.block_id,
        ).values_list('shared_with__username', flat=True)
        }
        return {
            'content': loader.render_template('templates/html/mentoring-table-shared-list.html', context)
        }

    @XBlock.json_handler
    def clear_notification(self, data, suffix=''):
        """
        Clear out notifications for users who shared with this user on the last page load.
        Since more users might share with them while they're viewing the page, only remove the ones
        that they had at the time.
        """
        usernames = data.get('usernames')
        if not usernames:
            raise JsonHandlerError(400, "No usernames sent.")
        try:
            isinstance(usernames, list)
        except ValueError:
            raise JsonHandlerError(400, "Usernames must be a list.")
        Share.objects.filter(
            shared_with__username=self.current_user_key,
            shared_by__username__in=usernames,
            block_id=self.block_id,
        ).update(
            notified=True
        )

    @property
    def block_id(self):
        usage_id = self.scope_ids.usage_id
        if isinstance(usage_id, six.string_types):
            return usage_id
        try:
            return six.text_type(usage_id.replace(branch=None, version_guid=None))
        except AttributeError:
            pass
        return six.text_type(usage_id)

    @XBlock.json_handler
    def share_results(self, data, suffix=''):
        target_usernames = data.get('usernames')
        target_usernames = [username.strip().lower() for username in target_usernames if username.strip()]
        current_user = User.objects.get(username=self.current_user_key)

        failed_users = []
        if not target_usernames:
            raise JsonHandlerError(400, _('Usernames not provided.'))
        for target_username in target_usernames:
            try:
                target_user = User.objects.get(username=target_username)
            except User.DoesNotExist:
                failed_users.append(target_username)
                continue
            if current_user == target_user:
                continue
            try:
                Share.objects.get(shared_by=current_user, shared_with=target_user, block_id=self.block_id)
            except Share.DoesNotExist:
                Share(
                    shared_by=current_user, submission_uid=self.runtime.anonymous_student_id, shared_with=target_user,
                    block_id=self.block_id,
                ).save()

        if failed_users:
            raise JsonHandlerError(
                400,
                _('Some users could not be shared with. Please check these usernames: {}').format(
                    ', '.join(failed_users)
                )
            )
        return {}

    @XBlock.json_handler
    def remove_share(self, data, suffix=''):
        target_username = data.get('username')
        if not target_username:
            raise JsonHandlerError(400, _('Username not provided.'))
        Share.objects.filter(
            shared_by__username=self.current_user_key,
            shared_with__username=target_username,
            block_id=self.block_id,
        ).delete()
        return {'message': _('Removed successfully.')}

    def student_view(self, context):
        context = context.copy() if context else {}
        fragment = Fragment()

        for child_id in self.children:
            child = self.runtime.get_block(child_id)
            # Child should be an instance of MentoringTableColumn
            child_frag = child.render('mentoring_view', context)
            fragment.add_frag_resources(child_frag)

        context['allow_sharing'] = self.allow_sharing
        context['allow_download'] = self.allow_download
        user_service = self.runtime.service(self, 'user')
        if user_service:
            context['view_options'] = Share.objects.filter(
                shared_with__username=self.current_user_key,
                block_id=self.block_id,
            ).values_list('shared_by__username', flat=True)
            context['username'] = self.current_user_key
            share_notifications = Share.objects.filter(
                shared_with__username=self.current_user_key,
                notified=False, block_id=self.block_id,
            ).values_list('shared_by__username', flat=True)
            context['share_notifications'] = share_notifications and json.dumps(list(share_notifications))

        if self.type:
            # Load an optional background image:
            context['bg_image_url'] = self.runtime.local_resource_url(self, 'public/img/{}-bg.png'.format(self.type))
            # Load an optional description for the background image, for accessibility
            try:
                context['bg_image_description'] = loader.load_unicode('static/text/table-{}.txt'.format(self.type))
            except IOError as e:
                if e.errno == errno.ENOENT:
                    pass
                else:
                    raise

        report_template = loader.render_template('templates/html/mentoring-table-report.html', {
            'title': self.display_name,
            'css': loader.load_unicode(AnswerRecapBlock.css_path) + loader.load_unicode(self.css_path),
            'student_name': self._get_user_full_name(),
            'course_name': self._get_course_name(),
        })

        fragment.add_content(loader.render_template('templates/html/mentoring-table-container.html', context))
        fragment.add_css_url(self.runtime.local_resource_url(self, 'public/css/mentoring-table.css'))
        fragment.add_javascript_url(self.runtime.local_resource_url(self, 'public/js/vendor/jquery-shorten.js'))
        fragment.add_javascript_url(self.runtime.local_resource_url(self, self.js_path))
        fragment.initialize_js(
            'MentoringTableBlock', {
                'reportContentSelector': '.mentoring-table-container',
                'reportTemplate': report_template,
            }
        )

        return fragment

    def mentoring_view(self, context):
        # Allow to render within mentoring blocks, or outside
        return self.student_view(context)

    def author_edit_view(self, context):
        """
        Add some HTML to the author view that allows authors to add choices and tips.
        """
        fragment = super(MentoringTableBlock, self).author_edit_view(context)
        fragment.add_content(loader.render_template('templates/html/mentoring-table-add-button.html', {}))
        # Share styles with the questionnaire edit CSS:
        fragment.add_css_url(self.runtime.local_resource_url(self, 'public/css/questionnaire-edit.css'))
        return fragment


class MentoringTableColumn(StudioEditableXBlockMixin, StudioContainerXBlockMixin, XBlock):
    """
    A column in a mentoring table. Has a header and can contain HTML and AnswerRecapBlocks.
    """
    display_name = String(display_name=_("Display Name"), default="Column")
    header = String(
        display_name=_("Header"),
        help=_("Header of this column"),
        default="",
        scope=Scope.content,
        multiline_editor="html",
    )
    editable_fields = ("header", )
    has_children = True

    def mentoring_view(self, context=None):
        """ Render this XBlock within a mentoring block. """
        context = context.copy() if context else {}
        fragment = Fragment()
        for child_id in self.children:
            child = self.runtime.get_block(child_id)
            if child.scope_ids.block_type == "html":
                # HTML block current doesn't support "mentoring_view" and if "student_view" is used, it gets wrapped
                # with HTML we don't want. So just grab its HTML directly.
                child_frag = Fragment(child.data)
            else:
                child_frag = child.render('mentoring_view', context)
            fragment.add_content(child_frag.content)
            fragment.add_frag_resources(child_frag)
        return fragment

    def author_preview_view(self, context):
        return self.mentoring_view(context)

    def student_view(self, context=None):
        """ Normal view of this XBlock, identical to mentoring_view """
        return self.mentoring_view(context)

    def author_edit_view(self, context):
        """
        Add some HTML to the author view that allows authors to add choices and tips.
        """
        fragment = super(MentoringTableColumn, self).author_edit_view(context)
        fragment.content = u"<div style=\"font-weight: bold;\">{}</div>".format(self.header) + fragment.content
        fragment.add_content(loader.render_template('templates/html/mentoring-column-add-button.html', {}))
        # Share styles with the questionnaire edit CSS:
        fragment.add_css_url(self.runtime.local_resource_url(self, 'public/css/questionnaire-edit.css'))
        return fragment



# -*- coding: utf-8 -*-
#
# Copyright (c) 2014-2015 Harvard, edX & OpenCraft
#
# This software's license gives you freedom; you can copy, convey,
# propagate, redistribute and/or modify this program under the terms of
# the GNU Affero General Public License (AGPL) as published by the Free
# Software Foundation (FSF), either version 3 of the License, or (at your
# option) any later version of the AGPL published by the FSF.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero
# General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program in a file in the toplevel directory called
# "AGPLv3".  If not, see <http://www.gnu.org/licenses/>.
#
"""
Test that the various title/display_name options for Answer and MCQ/MRQ/Ratings work.
"""

# Imports ###########################################################
import ddt
from mock import patch
from xblockutils.base_test import SeleniumXBlockTest

# Classes ###########################################################


@ddt.ddt
class TitleTest(SeleniumXBlockTest):
    """
    Test the various display_name/show_title options for Problem Builder
    """

    @ddt.data(
        ('<problem-builder show_title="false"><pb-answer name="a"/></problem-builder>', None),
        ('<problem-builder><pb-answer name="a"/></problem-builder>', "Problem Builder"),
        ('<problem-builder mode="assessment"><pb-answer name="a"/></problem-builder>', "Problem Builder"),
        ('<problem-builder display_name="A Question"><pb-answer name="a"/></problem-builder>', "A Question"),
        ('<problem-builder display_name="A Question" show_title="false"><pb-answer name="a"/></problem-builder>', None),
    )
    @ddt.unpack
    def test_title(self, xml, expected_title):
        self.set_scenario_xml(xml)
        pb_element = self.go_to_view()
        if expected_title is not None:
            h3 = pb_element.find_element_by_css_selector('h3')
            self.assertEqual(h3.text, expected_title)
        else:
            # No <h3> element should be present:
            all_h3s = pb_element.find_elements_by_css_selector('h3')
            self.assertEqual(len(all_h3s), 0)


class StepTitlesTest(SeleniumXBlockTest):
    """
    Test that the various title/display_name options for Answer and MCQ/MRQ/Ratings work.
    """

    test_parameters = (
        # display_name, show_title?, expected_title:  (None means default value)
        ("Custom Title", None,  "Custom Title",),
        ("Custom Title", True,  "Custom Title",),
        ("Custom Title", False, None),
        ("",             None,  "Question"),
        ("",             True,  "Question"),
        ("",             False, None),
    )

    mcq_template = """
        <problem-builder>
            <pb-mcq name="mcq_1_1" question="Who was your favorite character?"
              correct_choices="[gaius,adama,starbuck,roslin,six,lee]"
              {display_name_attr} {show_title_attr}
            >
                <pb-choice value="gaius">Gaius Baltar</pb-choice>
                <pb-choice value="adama">Admiral William Adama</pb-choice>
                <pb-choice value="starbuck">Starbuck</pb-choice>
                <pb-choice value="roslin">Laura Roslin</pb-choice>
                <pb-choice value="six">Number Six</pb-choice>
                <pb-choice value="lee">Lee Adama</pb-choice>
            </pb-mcq>
        </problem-builder>
    """

    mrq_template = """
        <problem-builder>
            <pb-mrq name="mrq_1_1" question="What makes a great MRQ?"
              ignored_choices="[1,2,3]"
              {display_name_attr} {show_title_attr}
            >
                <pb-choice value="1">Lots of choices</pb-choice>
                <pb-choice value="2">Funny choices</pb-choice>
                <pb-choice value="3">Not sure</pb-choice>
            </pb-mrq>
        </problem-builder>
    """

    rating_template = """
        <problem-builder>
            <pb-rating name="rating_1_1" question="How do you rate Battlestar Galactica?"
              correct_choices="[5,6]"
              {display_name_attr} {show_title_attr}
            >
                <pb-choice value="6">More than 5 stars</pb-choice>
            </pb-rating>
        </problem-builder>
    """

    long_answer_template = """
        <problem-builder>
            <pb-answer name="answer_1_1" question="What did you think of the ending?"
              {display_name_attr} {show_title_attr} />
        </problem-builder>
    """

    def setUp(self):
        super(StepTitlesTest, self).setUp()
        # Disable asides for this test since the acid aside seems to cause Database errors
        # When we test multiple scenarios in one test method.
        patcher = patch(
            'workbench.runtime.WorkbenchRuntime.applicable_aside_types',
            lambda self, block: [], create=True
        )
        patcher.start()
        self.addCleanup(patcher.stop)

    def test_all_the_things(self):
        """ Test various permutations of our problem-builder components and title options. """
        # We use a loop within the test rather than DDT, because this is WAY faster
        # since we can bypass the Selenium set-up and teardown
        for display_name, show_title, expected_title in self.test_parameters:
            for qtype in ("mcq", "mrq", "rating", "long_answer"):
                template = getattr(self, qtype + "_template")
                xml = template.format(
                    display_name_attr='display_name="{}"'.format(display_name) if display_name is not None else "",
                    show_title_attr='show_title="{}"'.format(show_title) if show_title is not None else "",
                )
                self.set_scenario_xml(xml)
                pb_element = self.go_to_view()
                if expected_title:
                    h4 = pb_element.find_element_by_css_selector('h4')
                    self.assertEqual(h4.text, expected_title)
                else:
                    # No <h4> element should be present:
                    all_h4s = pb_element.find_elements_by_css_selector('h4')
                    self.assertEqual(len(all_h4s), 0)



"""
Unit tests for DashboardVisualData
"""
import unittest

from problem_builder.dashboard_visual import DashboardVisualData


class TestDashboardVisualData(unittest.TestCase):
    """
    Test DashboardVisualData with some mocked data
    """
    def test_construct_data(self):
        """
        Test parsing of data and creation of SVG filter data.
        """
        blocks = [
            {
                'display_name': 'Block 1',
                'mcqs': [],
                'has_average': True,
                'average': 0,
            },
            {
                'display_name': 'Block 2',
                'mcqs': [],
                'has_average': True,
                'average': 1.3,
            },
            {
                'display_name': 'Block 3',
                'mcqs': [],
                'has_average': True,
                'average': 30.8,
            },
        ]
        rules = {
            "images": [
                "step1.png",
                "step2.png",
                "step3.png",
            ],
            "background": "background.png",
            "overlay": "overlay.png",
            "width": "500",
            "height": "500"
        }

        def color_for_value(value):
            """ Mock color_for_value """
            return "red" if value > 1 else None

        data = DashboardVisualData(blocks, rules, color_for_value, "Visual Repr", "Description here")
        self.assertEqual(len(data.layers), 5)
        self.assertEqual(data.layers[0]["url"], "background.png")
        self.assertEqual(data.layers[4]["url"], "overlay.png")
        self.assertEqual(data.width, 500)
        self.assertEqual(data.height, 500)
        # Check the three middle layers built from the average values:
        self.assertEqual(data.layers[1]["url"], "step1.png")
        self.assertEqual(data.layers[1].get("color"), None)

        self.assertEqual(data.layers[2]["url"], "step2.png")
        self.assertEqual(data.layers[2]["color"], "red")

        self.assertEqual(data.layers[3]["url"], "step3.png")
        self.assertEqual(data.layers[3]["color"], "red")



# -*- coding: utf-8 -*-
#
# Copyright (c) 2014-2015 Harvard, edX & OpenCraft
#
# This software's license gives you freedom; you can copy, convey,
# propagate, redistribute and/or modify this program under the terms of
# the GNU Affero General Public License (AGPL) as published by the Free
# Software Foundation (FSF), either version 3 of the License, or (at your
# option) any later version of the AGPL published by the FSF.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero
# General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program in a file in the toplevel directory called
# "AGPLv3".  If not, see <http://www.gnu.org/licenses/>.
#
"""
Instructor Tool: An XBlock for instructors to export student answers from a course.

All processing is done offline.
"""
import json

import six
from django.core.paginator import Paginator
from xblock.core import XBlock
from xblock.exceptions import JsonHandlerError
from xblock.fields import Dict, List, Scope, String
from xblock.fragment import Fragment
from xblockutils.resources import ResourceLoader

loader = ResourceLoader(__name__)

PAGE_SIZE = 15

# URL Path to the Course Blocks REST API.
# Note that we add a trailing slash to avoid the API's redirect hit.
COURSE_BLOCKS_API = '/api/courses/v1/blocks/'


# Make '_' a no-op so we can scrape strings
def _(text):
    return text


@XBlock.needs("i18n")
@XBlock.wants('user')
class InstructorToolBlock(XBlock):
    """
    InstructorToolBlock: An XBlock for instructors to export student answers from a course.

    All processing is done offline.
    """
    display_name = String(
        display_name=_("Title (Display name)"),
        help=_("Title to display"),
        default=_("Instructor Tool"),
        scope=Scope.settings
    )
    active_export_task_id = String(
        # The UUID of the celery AsyncResult for the most recent export,
        # IF we are sill waiting for it to finish
        default="",
        scope=Scope.user_state,
    )
    last_export_result = Dict(
        # The info dict returned by the most recent successful export.
        # If the export failed, it will have an "error" key set.
        default=None,
        scope=Scope.user_state,
    )
    display_data = List(
        # The list of results associated with the most recent successful export.
        # Stored separately to avoid the overhead of sending it to the client.
        default=None,
        scope=Scope.user_state,
    )
    has_author_view = True

    @property
    def display_name_with_default(self):
        return "Instructor Tool"

    def author_view(self, context=None):
        """ Studio View """
        # Warn the user that this block will only work from the LMS. (Since the CMS uses
        # different celery queues; our task listener is waiting for tasks on the LMS queue)
        return Fragment(u'<p>Instructor Tool Block</p><p>This block only works from the LMS.</p>')

    def studio_view(self, context=None):
        """ View for editing Instructor Tool block in Studio. """
        # Display friendly message explaining that the block is not editable.
        return Fragment(u'<p>This is a preconfigured block. It is not editable.</p>')

    def check_pending_export(self):
        """
        If we're waiting for an export, see if it has finished, and if so, get the result.
        """
        from .tasks import export_data as export_data_task  # Import here since this is edX LMS specific
        if self.active_export_task_id:
            async_result = export_data_task.AsyncResult(self.active_export_task_id)
            if async_result.ready():
                self._save_result(async_result)

    def _save_result(self, task_result):
        """ Given an AsyncResult or EagerResult, save it. """
        self.active_export_task_id = ''
        if task_result.successful():
            if isinstance(task_result.result, dict) and not task_result.result.get('error'):
                self.display_data = task_result.result['display_data']
                del task_result.result['display_data']
                self.last_export_result = task_result.result
            else:
                self.last_export_result = {'error': u'Unexpected result: {}'.format(repr(task_result.result))}
                self.display_data = None
        else:
            self.last_export_result = {'error': six.text_type(task_result.result)}
            self.display_data = None

    @XBlock.json_handler
    def get_result_page(self, data, suffix=''):
        """ Return requested page of `last_export_result`. """
        paginator = Paginator(self.display_data, PAGE_SIZE)
        page = data.get('page', None)
        return {
            'display_data': paginator.page(page).object_list,
            'num_results': len(self.display_data),
            'page_size': PAGE_SIZE
        }

    def student_view(self, context=None):
        """ Normal View """
        if not self.user_is_staff():
            return Fragment(u'<p>This interface can only be used by course staff.</p>')
        block_choices = {
            _('Multiple Choice Question'): 'MCQBlock',
            _('Multiple Response Question'): 'MRQBlock',
            _('Rating Question'): 'RatingBlock',
            _('Long Answer'): 'AnswerBlock',
        }

        html = loader.render_template('templates/html/instructor_tool.html', {
            'block_choices': block_choices,
            'course_blocks_api': COURSE_BLOCKS_API,
            'root_block_id': six.text_type(getattr(self.runtime, 'course_id', 'course_id')),
        })
        fragment = Fragment(html)
        fragment.add_css_url(self.runtime.local_resource_url(self, 'public/css/instructor_tool.css'))
        fragment.add_javascript_url(self.runtime.local_resource_url(self, 'public/js/instructor_tool.js'))
        fragment.add_javascript_url(self.runtime.local_resource_url(self, 'public/js/vendor/underscore-min.js'))
        fragment.add_javascript_url(self.runtime.local_resource_url(self, 'public/js/vendor/backbone-min.js'))
        fragment.add_javascript_url(self.runtime.local_resource_url(self, 'public/js/vendor/backbone.paginator.min.js'))
        fragment.initialize_js('InstructorToolBlock')
        return fragment

    @property
    def download_url_for_last_report(self):
        """ Get the URL for the last report, if any """
        # Unfortunately this is a bit inefficient due to the ReportStore API
        if not self.last_export_result or self.last_export_result['error'] is not None:
            return None
        from lms.djangoapps.instructor_task.models import ReportStore
        report_store = ReportStore.from_config(config_name='GRADES_DOWNLOAD')
        course_key = getattr(self.scope_ids.usage_id, 'course_key', None)
        return dict(report_store.links_for(course_key)).get(self.last_export_result['report_filename'])

    def _get_status(self):
        self.check_pending_export()
        return {
            'export_pending': bool(self.active_export_task_id),
            'last_export_result': self.last_export_result,
            'download_url': self.download_url_for_last_report,
        }

    def raise_error(self, code, message):
        """
        Raises an error and marks the block with a simulated failed task dict.
        """
        self.last_export_result = {
            'error': message,
        }
        self.display_data = None
        raise JsonHandlerError(code, message)

    @XBlock.json_handler
    def get_status(self, data, suffix=''):
        return self._get_status()

    @XBlock.json_handler
    def delete_export(self, data, suffix=''):
        self._delete_export()
        return self._get_status()

    def _delete_export(self):
        self.last_export_result = None
        self.display_data = None
        self.active_export_task_id = ''

    @XBlock.json_handler
    def start_export(self, data, suffix=''):
        """ Start a new asynchronous export """
        block_types = data.get('block_types', None)
        usernames = data.get('usernames', None)
        root_block_id = data.get('root_block_id', None)
        match_string = data.get('match_string', None)

        # Process user-submitted data
        if block_types == 'all':
            block_types = []
        else:
            block_types = [block_types]

        user_service = self.runtime.service(self, 'user')
        if not self.user_is_staff():
            return {'error': 'permission denied'}
        if not usernames:
            user_ids = None
        else:
            user_ids = []
            for username in usernames.split(','):
                username = username.strip()
                user_id = user_service.get_anonymous_user_id(username, six.text_type(self.runtime.course_id))
                if user_id:
                    user_ids.append(user_id)
            if not user_ids:
                self.raise_error(404, _("Could not find any of the specified usernames."))

        if not root_block_id:
            root_block_id = self.scope_ids.usage_id
            # Block ID not in workbench runtime.
            root_block_id = six.text_type(getattr(root_block_id, 'block_id', root_block_id))

        # Launch task
        from .tasks import export_data as export_data_task  # Import here since this is edX LMS specific
        self._delete_export()
        # Make sure we nail down our state before sending off an asynchronous task.
        self.save()
        async_result = export_data_task.delay(
            # course_id not available in workbench.
            six.text_type(getattr(self.runtime, 'course_id', 'course_id')),
            root_block_id,
            block_types,
            user_ids,
            match_string,
        )
        if async_result.ready():
            # In development mode, the task may have executed synchronously.
            # Store the result now, because we won't be able to retrieve it later :-/
            if async_result.successful():
                # Make sure the result can be represented as JSON, since the non-eager celery
                # requires that
                json.dumps(async_result.result)
            self._save_result(async_result)
        else:
            # The task is running asynchronously. Store the result ID so we can query its progress:
            self.active_export_task_id = async_result.id
        return self._get_status()

    @XBlock.json_handler
    def cancel_export(self, request, suffix=''):
        from .tasks import export_data as export_data_task  # Import here since this is edX LMS specific
        if self.active_export_task_id:
            async_result = export_data_task.AsyncResult(self.active_export_task_id)
            async_result.revoke()
            self._delete_export()

    def _get_user_attr(self, attr):
        """Get an attribute of the current user."""
        user_service = self.runtime.service(self, 'user')
        if user_service:
            # May be None when creating bok choy test fixtures
            return user_service.get_current_user().opt_attrs.get(attr)
        return None

    def user_is_staff(self):
        """Return a Boolean value indicating whether the current user is a member of staff."""
        return self._get_user_attr('edx-platform.user_is_staff')



"""
Unit tests for models.
"""
from django.test import TestCase
from mock import MagicMock, PropertyMock

from problem_builder.models import Answer, delete_anonymous_user_answers


class AnswerDeleteSignalTest(TestCase):
    """ Unit tests for pre_delete signal receiver. """

    def setUp(self):
        super(AnswerDeleteSignalTest, self).setUp()
        self.course_id = 'course-v1:edX+DemoX+Demo_Course'
        self.anonymous_student_id = '123456789876543210'
        Answer.objects.create(
            name='test-course-key',
            student_id=self.anonymous_student_id,
            course_key=self.course_id,
        )
        Answer.objects.create(
            name='test-course-key-2',
            student_id=self.anonymous_student_id,
            course_key=self.course_id,
        )
        Answer.objects.create(
            name='test-course-key',
            student_id='other-user',
            course_key=self.course_id,
        )

    def test_delete_anonymous_user_answers(self):
        """
        Test the receiver function by calling it directly.
        """
        anonymous_user_id_mock = MagicMock()
        type(anonymous_user_id_mock).anonymous_user_id = PropertyMock(return_value=self.anonymous_student_id)
        delete_anonymous_user_answers(MagicMock(), instance=anonymous_user_id_mock)
        self.assertEqual(Answer.objects.filter(student_id=self.anonymous_student_id).count(), 0)
        self.assertEqual(Answer.objects.exclude(student_id=self.anonymous_student_id).count(), 1)



# -*- coding: utf-8 -*-
#
# Copyright (c) 2014-2015 Harvard, edX & OpenCraft
#
# This software's license gives you freedom; you can copy, convey,
# propagate, redistribute and/or modify this program under the terms of
# the GNU Affero General Public License (AGPL) as published by the Free
# Software Foundation (FSF), either version 3 of the License, or (at your
# option) any later version of the AGPL published by the FSF.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero
# General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program in a file in the toplevel directory called
# "AGPLv3".  If not, see <http://www.gnu.org/licenses/>.
#
"""
The mentoring XBlock was previously designed to be edited using XML.

This file contains a script to help migrate mentoring blocks to the new format which is
optimized for editing in Studio.

To run the script on devstack:
SERVICE_VARIANT=cms DJANGO_SETTINGS_MODULE="cms.envs.devstack" python -m problem_builder.v1.upgrade [course id here]

You can add "--version=v0" at the end of the command to upgrade from the oldest xblock-mentoring
instance which was kept on gsehub's GitHub account.
"""
import logging
import sys
import warnings

import six
from lxml import etree
from six import StringIO

from courseware.models import StudentModule
from mentoring import MentoringBlock
from problem_builder.mentoring import MentoringBlock as NewMentoringBlock

from .studio_xml_utils import studio_update_from_node
from .xml_changes import convert_xml_to_v2


def upgrade_block(store, block, from_version="v1"):
    """
    Given a MentoringBlock "block" with old-style (v1) data in its "xml_content" field, parse
    the XML and re-create the block with new-style (v2) children and settings.
    """
    assert isinstance(block, (MentoringBlock, NewMentoringBlock))
    assert bool(block.xml_content)  # If it's a v1 block it will have xml_content
    xml_content_str = block.xml_content
    parser = etree.XMLParser(remove_blank_text=True)
    root = etree.parse(StringIO(xml_content_str), parser=parser).getroot()
    assert root.tag == "mentoring"
    with warnings.catch_warnings(record=True) as warnings_caught:
        warnings.simplefilter("always")
        convert_xml_to_v2(root, from_version=from_version)
        for warning in warnings_caught:
            print(u"    ➔ {}".format(six.text_type(warning.message)))

    # We need some special-case handling to deal with HTML being an XModule and not a pure XBlock:
    try:
        from xmodule.html_module import HtmlDescriptor
    except ImportError:
        pass  # Perhaps HtmlModule has been converted to an XBlock?
    else:
        @classmethod
        def parse_xml_for_HtmlDescriptor(cls, node, runtime, keys, id_generator):
            block = runtime.construct_xblock_from_class(cls, keys)
            block.data = node.text if node.text else ""
            for child in list(node):
                if isinstance(child.tag, six.string_types):
                    block.data += etree.tostring(child)
            return block
        HtmlDescriptor.parse_xml = parse_xml_for_HtmlDescriptor

    # Save the xml_content to make this processes rerunnable, in case it doesn't work correctly the first time.
    root.attrib["xml_content"] = xml_content_str

    # Was block already published?
    parent = store.get_item(block.parent)  # Don't use get_parent()/get_block() as it may be an outdated cached version
    assert getattr(parent.location, 'branch', None) is None
    assert getattr(parent.location, 'version_guid', None) is None
    assert getattr(parent.children[0], 'branch', None) is None
    assert getattr(parent.children[0], 'version_guid', None) is None
    parent_was_published = not store.has_changes(parent)

    old_usage_id = block.location
    if old_usage_id.block_type != "problem-builder":
        # We need to change the block type from "mentoring" to "problem-builder", which requires
        # deleting then re-creating the block:
        store.delete_item(old_usage_id, user_id=None)
        parent_children = parent.children
        index = parent_children.index(old_usage_id)

        url_name = six.text_type(old_usage_id.block_id)
        if "url_name" in root.attrib:
            url_name_xml = root.attrib.pop("url_name")
            if url_name != url_name_xml:
                print(u"    ➔ Two conflicting url_name values! Using the 'real' one : {}".format(url_name))
                print(u"    ➔ References to the old url_name ({}) need to be updated manually.".format(url_name_xml))
        block = store.create_item(
            user_id=None,
            course_key=old_usage_id.course_key,
            block_type="problem-builder",
            block_id=url_name,
            fields={"xml_content": xml_content_str},
        )
        parent_children[index] = block.location
        parent.save()
        store.update_item(parent, user_id=None)
        print(u"    ➔ problem-builder created: {}".format(url_name))

        # Now we've changed the block's block_type but in doing so we've disrupted the student data.
        # Migrate it now:
        student_data = StudentModule.objects.filter(module_state_key=old_usage_id)
        num_entries = student_data.count()
        if num_entries > 0:
            print(u"    ➔ Migrating {} student records to new block".format(num_entries))
            student_data.update(module_state_key=block.location)

    # Replace block with the new version and the new children:
    studio_update_from_node(block, root)

    if parent_was_published:
        store.publish(parent.location, user_id=None)


if __name__ == '__main__':
    # Disable some distracting overly-verbose warnings that we don't need:
    for noisy_module in ('edx.modulestore', 'elasticsearch', 'urllib3.connectionpool'):
        logging.getLogger(noisy_module).setLevel(logging.ERROR)
    from opaque_keys import InvalidKeyError
    from opaque_keys.edx.keys import CourseKey
    from xmodule.modulestore.django import modulestore

    print(u"┏━━━━━━━━━━━━━━━━━━━━━━━━━━┓")
    print(u"┃ Mentoring Upgrade Script ┃")
    print(u"┗━━━━━━━━━━━━━━━━━━━━━━━━━━┛")

    try:
        course_id = CourseKey.from_string(sys.argv[1])
    except (IndexError, InvalidKeyError):
        sys.exit("Need a course ID argument like 'HarvardX/GSE1.1x/3T2014' or 'course-v1:HarvardX+B101+2015'")

    from_version = "v1"
    if len(sys.argv) > 2:
        if sys.argv[2] == "--version=v0":
            from_version = "v0"
        else:
            sys.exit("invalid second argument: {}".format(' '.join(sys.argv[2:])))

    store = modulestore()
    course = store.get_course(course_id)
    if course is None:
        sys.exit(u"Course '{}' not found.".format(six.text_type(course_id)))
    print(u" ➔ Found course: {}".format(course.display_name))
    print(u" ➔ Searching for mentoring blocks")
    blocks_found = []

    def find_mentoring_blocks(block):
        """
        Find mentoring and recently-upgraded blocks. We check the recently upgraded ones
        in case an error happened and we're re-running the upgrade.
        """
        # If it's a v1 block or a recently upgraded block it will have xml_content:
        if isinstance(block, (MentoringBlock, NewMentoringBlock)) and block.xml_content:
            blocks_found.append(block.scope_ids.usage_id)
        elif block.has_children:
            for child_id in block.children:
                find_mentoring_blocks(block.runtime.get_block(child_id))
    find_mentoring_blocks(course)

    total = len(blocks_found)
    print(u" ➔ Found {} mentoring blocks".format(total))

    print(u" ➔ Doing a quick sanity check of the url_names")
    url_names = set()
    stop = False
    for block_id in blocks_found:
        url_name = block_id.block_id
        block = course.runtime.get_block(block_id)
        if url_name in url_names:
            print(u" ➔ Mentoring block {} appears in the course in multiple places!".format(url_name))
            print(u'   (display_name: "{}", parent {}: "{}")'.format(
                block.display_name, block.parent, block.get_parent().display_name
            ))
            print(u'   To fix, you must delete the extra occurences.')
            stop = True
            continue
        if block.url_name and block.url_name != six.text_type(block_id.block_id):
            print(u" ➔ Warning: Mentoring block {} has a different url_name set in the XML.".format(url_name))
            print(u"   If other blocks reference this block using the XML url_name '{}',".format(block.url_name))
            print(u"   those blocks will need to be updated.")
            if "--force" not in sys.argv:
                print(u"   In order to force this upgrade to continue, add --force to the end of the command.")
                stop = True
        url_names.add(url_name)

    if stop:
        sys.exit(u" ➔ Exiting due to errors preventing the upgrade.")

    with store.bulk_operations(course.location.course_key):
        count = 1
        for block_id in blocks_found:
            block = course.runtime.get_block(block_id)
            print(u" ➔ Upgrading block {} of {} - \"{}\"".format(count, total, block.url_name))
            count += 1
            upgrade_block(store, block, from_version)

    print(u" ➔ Complete.")
